#%%
# import os
# os.environ["RUST_BACKTRACE"] = "1"

import uuid
from typing import Optional, Tuple

import pytest
import torch
from torch.testing import assert_close

import rust_circuit as rc
from rust_circuit.py_utils import I

#%%

# order of importance:

# this
# get_update
# pdi
# nest
# (other test files, maybe too)

# %% [markdown]

# # Basic intro

# In this notebook, we'll demo generalfunctions and show how you can extend
# circuits by implementing your own generalfunction.
#
# You should think of this as an extension - if you
# implement the spec incorrectly this can cause pancis or bugs (we might
# improve this later).
#
# That said, there are nice functions for automatically
# testing your generalfunction specs for the properties we'll assume.

# %%

inp = rc.Array.randn(3, 4, 5)

# we can call functions like `softmax`
sm = rc.softmax(inp)
assert_close(sm.evaluate(), torch.softmax(inp.value, dim=-1))

# we can also do this (more verbosely) via 'new_by_name'
verbose_sm = rc.GeneralFunction.new_by_name(inp, spec_name="softmax")
assert sm == verbose_sm

# to see the full list of functions like this, goto definition on softmax
# (which will take you to the stub file for the library)

# %%

# functions will fail to be constructed if inputs shapes aren't viable.
with pytest.raises(rc.GeneralFunctionShapeNDimTooSmallError) as ex:
    rc.softmax(rc.Array.randn())
print(ex.exconly())

# %%

# other than these 'simple' general functions, the only other function we
# currently have implemented is `index_gen`. See interp/circuit/interop_rust/test_gen_index.py
# for examples.

# %%

# we can define additional simple functions like this via subclassing
# rc.GeneralFunctionSpecBase (we could add helpers for simple usecases if this
# becomes common, but hopefully we'll just upstream most simple functions?)


class MyCustomSigmoid(rc.GeneralFunctionSpecBase):
    def compute_hash_bytes(self) -> bytes:
        # IMPORTANT NOTE: make sure you update the hash if you're iterating on a general function in a notebook.
        # If you don't, then caches might be invalid and you may get an old version of the generalfunction!
        #
        # Hash is computed by default from class name + module name + method bytecode, should usually be right
        return uuid.UUID("36a3ef34-a3cb-49a5-8cae-89f29f269d78").bytes

    def function(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        return torch.sigmoid(x)

    def get_shape_info(self, *shapes) -> rc.GeneralFunctionShapeInfo:
        return rc.get_shape_info_simple(shapes)

    # optionally you can implement `get_device_dtype_override`, which validates input device/dtypes and optionally returns the output device/dtype
    # NOTE: if you do fancy stuff here, you'll need to manage dtype device upcasting! (TODO: better docs on this)
    def get_device_dtype_override(self, *device_dtypes: rc.TorchDeviceDtypeOp) -> Optional[rc.TorchDeviceDtypeOp]:
        assert device_dtypes[0].dtype in {
            "float16",
            "float32",
            "float64",
            "bfloat16",
        }  # softmax only applies to floats, not ints
        return None  # output has same device/dtype as input


# %%

func = MyCustomSigmoid.new(inp)
assert_close(func.evaluate(), torch.sigmoid(inp.value))
func

# %%

# this function will support adding extra batch dimensions:
bigger_inp = rc.Array.randn(7, *inp.shape)
bigger_func = rc.Expander((inp, lambda _: bigger_inp))(func)
assert_close(bigger_func.evaluate(), torch.sigmoid(bigger_inp.value))
bigger_func

# %%

# this custom function doesn't support bijection
with pytest.raises(RuntimeError):
    func.print(rc.PrintOptions(bijection=True))
# but support usual printing
func.print(rc.PrintOptions(bijection=False))

# %%
# but we support bijection for custom functions stored in rrfs
# NOTE: python caches imports, so editing the file and rerunning won't work in a notebook!
func_stored_in_rrfs = rc.GeneralFunction.new_by_path(inp, path="/test/general_function.py:MyCustomSigmoid")
func_stored_in_rrfs.print(rc.PrintOptions(bijection=True))

# %%

# now let's define a function which has more interesting batching properties
class MyCustomSoftmaxAndSum(rc.GeneralFunctionSpecBase):
    def function(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        return torch.softmax(x, dim=-2).sum(dim=-1)

    def get_shape_info(self, *shapes: rc.Shape) -> rc.GeneralFunctionShapeInfo:
        # here, we set some batching options. See GeneralFunctionShapeInfo docs for what this does!
        #
        # this can throw an exception if the shapes are invalid.
        #
        # note this is just a helper for simple cases - we can do whatever (valid) shape specification we want
        # TODO: these docs could be improved
        return rc.get_shape_info_simple(shapes, num_non_batchable_output_dims=1, removed_from_end=1)


# %%

new_func = MyCustomSoftmaxAndSum.new(inp)
assert_close(new_func.evaluate(), torch.softmax(inp.value, dim=-2).sum(dim=-1))
new_func

# %%

# we can test this has the advertized batching properties:
# see GeneralFunctionSpecTester docs for more details
rc.GeneralFunctionSpecTester(test_with_rand=True).test_many_shapes(MyCustomSoftmaxAndSum())
# note that this just checks that our spec is consistent and the function
# respects the shape info - not anything else.
# For instance, if your 'get_shape_info' always errors, that will pass as it's consistent behavior.
#
# Note that if your spec doesn't have these consistency properties, 'rust_circuit' might panic!
# (That's part of why we provide an automatic test suite :) )
# (This isn't that bad, mostly just worse error messages)

# %%

# Note that this function doesn't support batching over later dims.
# this means (for instance) that we can't push down an index through these dims.

indexed = new_func.index(I[:, 1])
with pytest.raises(rc.PushDownIndexError) as ex1:
    rc.push_down_index(indexed)
print(ex1.exconly())

# but we can push down index through first dim
other_indexed = new_func.index(I[1])
pushed = rc.push_down_index(other_indexed)
assert_close(pushed.evaluate(), torch.softmax(inp.value, dim=-2).sum(dim=-1)[1])
pushed

# %%

# suppose we defined the same function but buggy
class MyCustomSoftmaxAndSumBuggy(MyCustomSoftmaxAndSum):
    def get_shape_info(self, *shapes: rc.Shape) -> rc.GeneralFunctionShapeInfo:
        # this dim actually isn't batchable!
        return rc.get_shape_info_simple(shapes, num_non_batchable_output_dims=0, removed_from_end=1)


# %%

# then our test will fail (in this case, the test fails because our spec claims
# that we could operate on inputs with ndim=1, but we actually can't and this
# wil fail with dimension out of range!)
with pytest.raises((rc.ExceptionWithRustContext, IndexError)) as ex2:
    rc.GeneralFunctionSpecTester(test_with_rand=True).test_many_shapes(MyCustomSoftmaxAndSumBuggy())
print(ex2.exconly())

# %%

# here we have a function which actually can operate on an input of any shape, but it actually doesn't support batching at all!
class BuggyMaxThenExpand(rc.GeneralFunctionSpecBase):
    def function(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        if x.numel() == 0:
            return torch.full_like(x, 0.0)
        return torch.full_like(x, x.max().item())

    def get_shape_info(self, *shapes) -> rc.GeneralFunctionShapeInfo:
        return rc.get_shape_info_simple(shapes, num_non_batchable_output_dims=1)


# %%

with pytest.raises((rc.ExceptionWithRustContext, AssertionError)) as ex3:
    rc.GeneralFunctionSpecTester(test_with_rand=True).test_many_shapes(BuggyMaxThenExpand())
print(ex3.exconly())

# %%


# let's do a function with fancier shape related properties
class FancyShape(rc.GeneralFunctionSpecBase):
    correct_x_thingy: int

    def __init__(self, correct_x_thingy: int) -> None:
        super().__init__()  # doesn't do anything atm
        self.correct_x_thingy = correct_x_thingy

    @property
    def name(self) -> str:
        return f"fancy_shape_with={self.correct_x_thingy}"

    def function(self, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        # random function with appropriate shape stuff
        [x, y, z] = rc.upcast_tensor_device_dtypes(
            [x, y, z]
        )  # we need this for properly managing scalars and none-dtype in general...
        # TODO: more docs on upcast_tensor_device_dtypes and what general functions have to do
        return torch.log_softmax(torch.einsum("... a b, c, ... -> ... b c", x, y, z), dim=-1)

    def get_shape_info(self, x_shape: rc.Shape, y_shape: rc.Shape, z_shape: rc.Shape) -> rc.GeneralFunctionShapeInfo:  # type: ignore[override]
        assert len(x_shape) >= 2
        assert len(y_shape) == 1
        assert x_shape[-1] == self.correct_x_thingy, f"last dim != thingy={self.correct_x_thingy}"

        return rc.GeneralFunctionShapeInfo(
            z_shape + (x_shape[-1],) + y_shape,
            num_non_batchable_output_dims=2,
            input_batchability=[True, False, True],
        )

    @classmethod
    def new(cls, x: rc.Circuit, y: rc.Circuit, z: rc.Circuit, name: Optional[str] = None):
        """convenience function"""
        return rc.GeneralFunction(x, y, z, spec=cls(7), name=name)


# %%

batch_shape: Tuple[int, ...] = (1, 3, 4)
x = rc.Array.randn(*batch_shape, 9, 7)
y = rc.Array.randn(8)
z = rc.Array.randn(*batch_shape)
FancyShape.new(x, y, z)

# %%

batch_shape: Tuple[int, ...] = (1, 3, 4)
x = rc.Array.randn(*batch_shape, 9, 7)
y = rc.Array.randn(8)
z = rc.Scalar(value=384.0, shape=batch_shape)
# this would fail if we didn't upcast tensors due to float64 vs float32
# (ok, technically einsum apparently is fine with differnet dtypes (!?), but in theory)
_ = FancyShape.new(x, y, z).evaluate()


# %%

x_new = rc.Array.randn(*batch_shape, 9, 8)
with pytest.raises((rc.ExceptionWithRustContext, AssertionError)) as ex4:
    FancyShape.new(x_new, y, z)
print(ex4.exconly())

# %%

# torch.einsum w/ opt_einsum doesn't support 0 dims & GeneralFunctionSpecTester generates them
torch.backends.opt_einsum.enabled = False  # type: ignore

# by default, this function will basically never find any valid shapes! We
# check that we do sometimes find valid shapes, so this will error (see `min_frac_successful` below)
with pytest.raises(RuntimeError) as ex8:
    rc.GeneralFunctionSpecTester(
        test_with_rand=True,
        samples_per_batch_dims=10,
        base_shapes_samples=1000,
        start_num_inputs=2,
        end_num_inputs=4,
        # this variable can be configured to change the minimum fraction of tests which do anything
        # see docs for more details
        min_frac_successful=0.1,
    ).test_many_shapes(FancyShape(5))
print(ex8.exconly())

# %%

# but we can run a custom generation scheme which gets more hits
for _ in range(1_000):
    batch_shape = tuple(int(x) for x in torch.randint(0, 7, size=(int(torch.randint(0, 4, ())),)))
    rc.GeneralFunctionSpecTester(test_with_rand=True, samples_per_batch_dims=10).test_from_shapes(
        FancyShape(5),
        [(*batch_shape, int(torch.randint(0, 7, ())), 5), (int(torch.randint(0, 7, ())),), batch_shape],
        shapes_must_be_valid=True,
    )

# %%


class FancyShapeBugged1(FancyShape):
    def get_shape_info(  # type: ignore[override]
        self, x_shape: rc.Shape, y_shape: rc.Shape, z_shape: rc.Shape
    ) -> rc.GeneralFunctionShapeInfo:
        out = super().get_shape_info(x_shape, y_shape, z_shape)
        out.num_non_batchable_output_dims = 1
        return out


with pytest.raises(RuntimeError) as ex5:
    for _ in range(1_000):
        batch_shape = tuple(int(x) for x in torch.randint(0, 7, size=(int(torch.randint(0, 4, ())),)))
        rc.GeneralFunctionSpecTester(test_with_rand=True, samples_per_batch_dims=10).test_from_shapes(
            FancyShapeBugged1(5),
            [(*batch_shape, int(torch.randint(0, 7, ())), 5), (int(torch.randint(0, 7, ())),), batch_shape],
            shapes_must_be_valid=True,
        )

print(ex5.exconly())

# %%


class FancyShapeBugged2(FancyShape):
    def function(self, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        out = super().function(x, y, z)
        if out.ndim >= 3:
            return torch.log_softmax(out, dim=-3)  # not actually batchable!
        return out


with pytest.raises((rc.ExceptionWithRustContext, AssertionError)) as ex6:
    for _ in range(1_000):
        batch_shape = tuple(int(x) for x in torch.randint(0, 7, size=(int(torch.randint(0, 4, ())),)))
        rc.GeneralFunctionSpecTester(test_with_rand=True, samples_per_batch_dims=10).test_from_shapes(
            FancyShapeBugged2(5),
            [(*batch_shape, int(torch.randint(0, 7, ())), 5), (int(torch.randint(0, 7, ())),), batch_shape],
            shapes_must_be_valid=True,
        )

print(ex6.exconly())

# %%



import torch

from rust_circuit import *


def test_save():
    tensor = torch.randn(3, 3)
    save_tensor(tensor)
    prefix = Array(tensor).tensor_hash_base16()
    loaded = get_tensor_prefix(prefix)
    print(tensor, loaded, torch.allclose(tensor, loaded))
    sync_all_unsynced_tensors()


if __name__ == "__main__":
    sync_all_unsynced_tensors()
    # test_save()



# %%
import uuid

import pytest
import torch

import rust_circuit as rc
import rust_circuit.module_library as mod_l
from rust_circuit.module_utils import drop_module_inputs

# %% [markdown]

# # Overview

# this notebook will go through:
# - Symbols
# - Expanding
# - Symbolic sizes
# - And how all of these things combing to yield Modules

# For demonstration purposes, we'll do various things with Expander which you
# should probably use Modules for in general. So, don't assume the stuff we do
# with 'Expander' is how you should do things. Similarly, we'll use some more
# verbose module construction approaches when you'll often want to use more
# consise approaches.

# We assume your're familiar with the basics of circuits + printing/parsing
# (see interp/demos/rust_circuit/printing_parsing.py for more on that)

# %% [markdown]

# # Symbols

# %%

# symbol nodes can be used as place holders in various cases
sym = rc.Symbol.new_with_random_uuid((2, 3), name="sym")
sym

# %%

# symbols have uuids. If the uuid is randomly generated, then the symbol is universally unique
sym.uuid

# %%

# symbols aren't explicitly computable
assert sym.is_explicitly_computable is False
with pytest.raises(rc.TensorEvalNotExplicitlyComputableError) as exc:
    sym.evaluate()
print(exc.exconly())

# %%

P = rc.Parser()

# if you parse a symbol without a uuid present, the null uuid will be used
null_uuid = P("'sym_with_random_uuid' [7] Symbol").cast_symbol().uuid
assert null_uuid == uuid.UUID("00000000-0000-0000-0000-000000000000")

# Ofc, you can specify a specific uuid. I recommed having some convenient way to generate new random uuids
# (For instance, I have a vim snippet which expands 'uuid' to a new uuid).

rand_uuid = P("'sym_with_random_uuid' [7] Symbol dc87d8f0-8d11-43d6-86ca-84bc097fae82").cast_symbol().uuid
rand_uuid

# %%

# you can also use 'rand'. But be warned, this will generate a new uuid each time the string is parsed.
rand_uuid = P("'sym_with_random_uuid' [7] Symbol rand").cast_symbol().uuid
rand_uuid

# %% [markdown]

# # Expanding
#
# Now let's talk about replacing symbols inside of circuits

# %%

# we can construct a circuit with symbols as placeholders
circ = P(
    """
'symbolic_mul' Einsum ij,jk->ik
  'my_sym' [7, 2] Symbol 0dc30ed9-239f-459d-869a-78c735e17b36
  'mat' [2, 5] Array rand
"""
)

printer = rc.PrintOptions(shape_only_when_necessary=False, colorer=rc.PrintOptions.type_colorer())

# this isn't possible to compute, but we can substitute in a value for sym and then compute
circ_comp = rc.Updater(lambda _: rc.Scalar(0.7, (7, 2), "rep"))(circ, "my_sym", fancy_validate=True)
print()
printer.print(circ_comp)
assert circ_comp.is_explicitly_computable

# %%

# often, when doing substitutions like this, we'd like to be able to add extra
# dimensions (which are batched over)
circ_batched = rc.Expander(("my_sym", lambda _: rc.Scalar(0.7, (9, 10, 11, 7, 2), "rep")))(circ, fancy_validate=True)
assert circ_batched.shape == (9, 10, 11, 7, 5)
printer.print(circ_batched)

# %%

# expanding also supports changing non-batched dims which can be changed
# (at least in most/many cases when this can be supported it is supported)

# we can change this 7 dim to a 12
circ_batched = rc.Expander(("my_sym", lambda _: rc.Scalar(0.7, (11, 12, 2), "rep")))(circ, fancy_validate=True)
assert circ_batched.shape == (11, 12, 5)
printer.print(circ_batched)


# but we can't change the 2 dim because it needs to correspond to 'mat'
with pytest.raises(rc.ConstructEinsumAxisSizeDifferentError) as exc0:
    rc.Expander(("my_sym", lambda _: rc.Scalar(0.7, (11, 12, 3), "rep")))(circ, fancy_validate=True)
print(exc0.exconly())

# %%

# for a more concrete example, let's consider a symbolic bilinear_mlp
# We'll need to set a few specific sizes

HIDDEN = 256
OUT_HIDDEN = 256
MLP_PROJ = 512
mlp = P(
    f"""
'bilinear_mlp' Einsum h,h,oh->o
  'bilinear_mlp.pre0' Index [0,:]
    'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
      'bilinear_mlp.pre' Einsum i,hi->h
        'bilinear_mlp.input' [{HIDDEN}] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
        'bilinear_mlp.w.proj_in' [{MLP_PROJ*2}, {HIDDEN}] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
  'bilinear_mlp.pre1' Index [1,:]
    'bilinear_mlp.fold_pre'
  'bilinear_mlp.w.proj_out' [{OUT_HIDDEN}, {MLP_PROJ}] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
"""
)

printer.print(mlp)

# %%

BATCH_SHAPE = (3, 5, 7)
NEW_HIDDEN = 17
NEW_OUT_HIDDEN = 13
NEW_MLP_PROJ = 11
mlp_expanded = rc.Expander(
    ("bilinear_mlp.input", lambda _: rc.Scalar(0.7, (*BATCH_SHAPE, NEW_HIDDEN), "rep_input")),
    ("bilinear_mlp.w.proj_in", lambda _: rc.Scalar(0.7, (NEW_MLP_PROJ * 2, NEW_HIDDEN), "rep_proj_in")),
    ("bilinear_mlp.w.proj_out", lambda _: rc.Scalar(0.7, (NEW_OUT_HIDDEN, NEW_MLP_PROJ), "rep_proj_out")),
)(mlp, fancy_validate=True)

print()
printer.print(mlp_expanded)

# %%

# Here's something which is a bit sad: we can't partially replace symbols with different sizes

with pytest.raises(rc.ConstructEinsumAxisSizeDifferentError) as excinfo:
    rc.Expander(
        ("bilinear_mlp.input", lambda _: rc.Scalar(0.7, (*BATCH_SHAPE, NEW_HIDDEN), "rep_input")),
    )(mlp, fancy_validate=True)
print(excinfo.exconly())

# to resolve this, we'll move onto the next section, symbolic sizes

# %% [markdown]

# # Symbolic sizes
#
# There's support for 'symbolic sizes' which can be resolved to different sizes without an error.
# If an impossible set of size assignments is detected, this will error.
# These are global in some important sense - they must be resolvable to the same size throughout a circuit.

# %%

sym_size_mlp = P(
    f"""
'bilinear_mlp' Einsum h,h,oh->o
  'bilinear_mlp.pre0' Index [0,:]
    'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
      'bilinear_mlp.pre' Einsum i,hi->h
                            # symbolic sizes can be indicated with <NUM>s
        'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
                                 # we can also indicate products (including products of symbolic sizes)
        'bilinear_mlp.w.proj_in' [?, ?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
  'bilinear_mlp.pre1' Index [1,:]
    'bilinear_mlp.fold_pre'
  'bilinear_mlp.w.proj_out' [?, ?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
"""
)

printer.print(sym_size_mlp)

# %%

# now we can partially replace!
sym_size = rc.Expander(
    ("bilinear_mlp.input", lambda _: rc.Scalar(0.7, (*BATCH_SHAPE, NEW_HIDDEN), "rep_input")),
)(sym_size_mlp, fancy_validate=True)
printer.print(sym_size)

# %%

# quick test of printing/parsing
parsed = P(
    """
'bilinear_mlp' Einsum pqrh,pqrh,oh->pqro
  'bilinear_mlp.pre0' Index [:,:,:,0,:]
    'bilinear_mlp.fold_pre' Rearrange c d e (a:2 b) -> c d e a:2 b
      'bilinear_mlp.pre' Einsum jkli,hi->jklh
        'rep_input' [3,5,7,17] Scalar 0.7
        'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
  'bilinear_mlp.pre1' Index [:,:,:,1,:]
    'bilinear_mlp.fold_pre'
  'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
"""
)
assert rc.deep_normalize(parsed) == rc.deep_normalize(sym_size)

# %%

sym_size_out = rc.Expander(
    ("bilinear_mlp.w.proj_out", lambda _: rc.Scalar(0.7, (NEW_OUT_HIDDEN, NEW_MLP_PROJ), "rep_proj_out"))
)(sym_size, fancy_validate=True)
printer.print(sym_size_out)

# %%

# we can also do the other one
sym_size_in = rc.Expander(
    ("bilinear_mlp.w.proj_in", lambda _: rc.Scalar(0.7, (NEW_MLP_PROJ * 2, NEW_HIDDEN), "rep_proj_in")),
)(sym_size, fancy_validate=True)
printer.print(sym_size_in)


# %%

# or finish substituting
sym_size_in_out = rc.Expander(
    ("bilinear_mlp.w.proj_out", lambda _: rc.Scalar(0.7, (NEW_OUT_HIDDEN, NEW_MLP_PROJ), "rep_proj_out"))
)(sym_size_in, fancy_validate=True)
printer.print(sym_size_in_out)
assert sym_size_in_out.is_explicitly_computable
assert mlp_expanded == sym_size_in_out

# %%

# we'll get an error if we make shapes inconsistent (in this case an einsum error)
MIS_MATCH_HIDDEN = 298
with pytest.raises(rc.ConstructError) as exc1:
    rc.Expander(
        ("bilinear_mlp.w.proj_in", lambda _: rc.Scalar(0.7, (NEW_MLP_PROJ * 2, MIS_MATCH_HIDDEN), "rep_proj_in")),
    )(sym_size, fancy_validate=True)
print(exc1.exconly())

# %%

# you can also get symbolic constraint errors like this example
scl0, scl1, v, new0, new1 = P.parse_circuits(
    """
# Here we are parsing a file with multiple circuits into a list of circuits
1 [7] Scalar 1.3
3 [7,3] Scalar 1.5
0 Einsum i,i,ij,j->i
  1
  2 [7] Scalar 1.5
  3
  4 [3] Scalar 1.7
'new0' [?] Symbol
'new1' [?] Symbol
"""
)

# %% [markdown]

# # Modules
#
# TODO: improve this explanation as needed
#
# Modules are a circuit type which themselves encode a symbolic substitution (where substitution is done via expand as shown above).
# We'll go through examples in a second, but let's do a quick explanation first.
#
# If we imagine a circuit with symbols as having 'free variables', then Modules
# are a call site which binds (some of) those free variables to specific
# inputs.
#
# Note that modules don't bind symbols which are bound by a child module - we'll go through some examples of this.
#
# Modules are built of two parts:
# - a ModuleSpec
# - and inputs
#
# ModuleSpecs can be thought of as 'lambdas' analogously to Modules being call sites. They consist of:
# - A spec circuit (which has symbols/'free variables', aka the function body)
# - And a list of 'ModuleArgSpec' which specifies which symbol to substitute out as well as any restrictions on expansion.
#
# Now let's build some modules!

# %%

input_sym, proj_in_sym, proj_out_sym, mlp = P.parse_circuits(
    """
'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
'bilinear_mlp.w.proj_in' [?, ?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
'bilinear_mlp.w.proj_out' [?, ?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
'bilinear_mlp' Einsum h,h,oh->o
  'bilinear_mlp.pre0' Index [0,:]
    'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
      'bilinear_mlp.pre' Einsum i,hi->h
                            # symbolic sizes can be indicated with <NUM>s
        'bilinear_mlp.input'
        'bilinear_mlp.w.proj_in'
  'bilinear_mlp.pre1' Index [1,:]
    'bilinear_mlp.fold_pre'
  'bilinear_mlp.w.proj_out'
"""
)
input_sym, proj_in_sym, proj_out_sym = [x.cast_symbol() for x in [input_sym, proj_in_sym, proj_out_sym]]

# we'll start with the straightforward, but verbose construction strategy
# see docs for ModuleArgSpec as needed
input_arg = rc.ModuleArgSpec(input_sym, batchable=True, expandable=True, ban_non_symbolic_size_expand=True)
proj_in_arg = rc.ModuleArgSpec(proj_in_sym, batchable=False, expandable=True, ban_non_symbolic_size_expand=True)
proj_out_arg = rc.ModuleArgSpec(proj_out_sym, batchable=False, expandable=True, ban_non_symbolic_size_expand=True)

mlp_spec = rc.ModuleSpec(
    mlp, [input_arg, proj_in_arg, proj_out_arg], check_all_inputs_used=True, check_unique_arg_names=True
)

input_arr = rc.Array.randn(*BATCH_SHAPE, NEW_HIDDEN, name="b.input")
proj_in_arr = rc.Array.randn(2 * NEW_MLP_PROJ, NEW_HIDDEN, name="b.proj_in")
proj_out_arr = rc.Array.randn(NEW_OUT_HIDDEN, NEW_MLP_PROJ, name="b.proj_out")

BATCH_SHAPE = (3, 5, 7)
NEW_HIDDEN = 17
NEW_OUT_HIDDEN = 13
NEW_MLP_PROJ = 11

# Now this module binds the inputs (input_arr, proj_in_arr, proj_out_arr)
# to the arg spec symbols (input_sym, proj_in_sym, proj_out_sym)
mlp_call = rc.Module(
    mlp_spec,
    "b.bilinear_mlp",
    **{
        "bilinear_mlp.input": input_arr,
        "bilinear_mlp.w.proj_in": proj_in_arr,
        "bilinear_mlp.w.proj_out": proj_out_arr,
    },
)

# modules have children: spec_circuit, input_0 ! input_sym_0, input_1 ! input_sym_1,...
# where input_0 is bound to input_sym_0
#
# If ModuleArgSpec flags aren't the default, then the are printed like input_0 ! input_sym_0 ftt
# where t/f correspond to true and false
print()
printer.print(mlp_call)

# %%

# as per usual, parsing works on this representation
assert mlp_call == P(mlp_call.repr())

# %%

# this module is extentionally equal to the following substitution (done with expand)
printer.print(mlp_call.substitute())

# %%

# we could have also used the 'new_flat' factory (which uses the argument order)
flat_mlp_call = rc.Module.new_flat(
    mlp_spec,
    input_arr,
    proj_in_arr,
    proj_out_arr,
    name="b.bilinear_mlp",
)
assert flat_mlp_call == mlp_call

# %%

# it's ofc possible to define modules/spec which only partially bind.

mlp_weights_spec = rc.ModuleSpec(mlp, [proj_in_arg, proj_out_arg])
mlp_weights = rc.Module(
    mlp_weights_spec,
    "bw.bilinear_mlp",
    **{
        "bilinear_mlp.w.proj_in": proj_in_arr,
        "bilinear_mlp.w.proj_out": proj_out_arr,
    },
)
printer.print(mlp_weights)

# %%

# and then we can nest this module as well.
mlp_w_inp_spec = rc.ModuleSpec(mlp_weights, [input_arg])
mlp_nested_mod = rc.Module(
    mlp_w_inp_spec,
    "b.bilinear_mlp",
    **{
        "bilinear_mlp.input": input_arr,
    },
)
printer.print(mlp_nested_mod)

# %%

substitutor = rc.Updater(lambda x: x.cast_module().substitute())
sub_inner_only = substitutor(mlp_nested_mod, "bw.bilinear_mlp")
printer.print(sub_inner_only)

# %%

sub_outer_only = substitutor(mlp_nested_mod, "b.bilinear_mlp")
printer.print(sub_outer_only)

# %%

# we can also sub all
assert mlp_call.substitute() == rc.substitute_all_modules(mlp_nested_mod)

# %%

# NOTE: in some cases, substituting an inner module may require Expander due to symbolic sizes!

# %%

# this is typically the most consise way to construct a module from a circuit with free symbols. See docs for details.
mlp_call_bind = rc.module_new_bind(
    mlp,
    ("bilinear_mlp.input", input_arr),
    rc.BindItem("bilinear_mlp.w.proj_in", proj_in_arr, batchable=False),
    rc.BindItem("bilinear_mlp.w.proj_out", proj_out_arr, batchable=False),
    name="b.bilinear_mlp",
)
assert mlp_call_bind == mlp_call

# %%

# there are also some utils for getting free symbols and constructing like this
assert rc.get_free_symbols(mlp_call) == []
assert rc.get_free_symbols(mlp) == [input_sym, proj_in_sym, proj_out_sym]
assert rc.get_free_symbols(mlp_weights) == [input_sym]

# this will use the default values for ModuleArgSpec (all true)
with_all_free = rc.ModuleSpec.new_free_symbols(mlp)
flat_mod = rc.Module.new_flat(with_all_free, input_arr, proj_in_arr, proj_out_arr, name="b.bilinear_mlp")

# there's printing/parsing shorthand for the all true case for Modules: all_t (see top line in print)
print()
printer.print(flat_mod)

# %% [markdown]

# modules are nice because they allow you to have fixed copy of the function
# body instead of needing to duplicate it with different shapes.
# this can allow for stuff like:
# - module specific rewrites for a given spec (in the same way we have einsum or rearrange specfic rewrites)
# - reusing a spec circuit multiple times with different shapes in the same circuit

# %%

RUNNING_HIDDEN = 20
RUNNING_MLP_PROJ = 11

new_proj_in_arr = rc.Array.randn(2 * RUNNING_MLP_PROJ, NEW_OUT_HIDDEN, name="b_new.proj_in")
new_proj_out_arr = rc.Array.randn(RUNNING_HIDDEN, RUNNING_MLP_PROJ, name="b_new.proj_out")

mlp_2_layer = rc.module_new_bind(
    mlp,
    ("bilinear_mlp.input", mlp_call),  # using this circuit as input
    rc.BindItem("bilinear_mlp.w.proj_in", new_proj_in_arr, batchable=False),
    rc.BindItem("bilinear_mlp.w.proj_out", new_proj_out_arr, batchable=False),
    name="outer_b.bilinear_mlp",
)

# note how this has the same spec circuit operating on different shapes!
# and this spec circuit is only printed once
print()
printer.print(mlp_2_layer)

# %%

# we can do rewrites on specs (but note that indexing into symbolic sizes might not work as desired!)
perm_einsum = rc.Updater(lambda x: rc.nest_einsums(x.cast_einsum(), (2, 1, 0)))(mlp_2_layer, "bilinear_mlp")
printer.print(perm_einsum)

# %%

# we can also do rewrites on just one of the specs
outer_spec_matcher = rc.IterativeMatcher("outer_b.bilinear_mlp").spec_circuit_matcher()
perm_einsum_diff = outer_spec_matcher.update(
    mlp_2_layer, lambda x: rc.nest_einsums(x.cast_einsum(), (2, 1, 0)).rename("bilinear_mlp_permute_args")
)
printer.print(perm_einsum_diff)

# %%


# you might find it useful to rename all circuits on the path to a given circuit within a module spec

# within the spec, match all parents of bilinear_mlp.pre and that circuit itself
within_spec_path_to_matcher = outer_spec_matcher.chain(rc.Matcher.match_any_found("bilinear_mlp.pre"))

perm_inside_einsum = within_spec_path_to_matcher.update(mlp_2_layer, lambda x: x.rename(x.name + " perm"))  # do rename
perm_inside_einsum = outer_spec_matcher.chain("bilinear_mlp.pre perm").update(
    perm_inside_einsum,
    lambda x: rc.nest_einsums(x.cast_einsum(), (1, 0)),  # and then finally update the circuit inside the spec
)
printer.print(perm_inside_einsum)

# %%

# it's sometimes useful to conform all symbolic shapes to their actual shape
# this results in the module specs no longer being equal because they were called on different shapes
printer.print(rc.conform_all_modules(mlp_2_layer))

# %%

# another useful thing we can do is get bound sub circuits
# (by pushing down modules)
# NOTE: ModulePusher has quite complex handling around batching dims. I've
# tried to test this well, but I currently put reasonably high probability on
# this being buggy in complex batching cases! (it would likely panic if buggy)
outer_bound_pre = rc.ModulePusher().get_unique_push_down_modules(
    mlp_2_layer, rc.IterativeMatcher("outer_b.bilinear_mlp").spec_circuit_matcher().chain("bilinear_mlp.pre")
)
printer.print(outer_bound_pre)
assert outer_bound_pre.is_explicitly_computable
actual = rc.substitute_all_modules(mlp_2_layer).get_unique(
    rc.restrict(rc.IterativeMatcher("bilinear_mlp.pre"), end_depth=5)
)
torch.testing.assert_close(outer_bound_pre.evaluate(), actual.evaluate())

# %%

# or for the inner module
inner_bound_pre = rc.ModulePusher().get_unique_push_down_modules(
    mlp_2_layer, rc.IterativeMatcher("b.bilinear_mlp").spec_circuit_matcher().chain("bilinear_mlp.pre")
)
printer.print(inner_bound_pre)
assert inner_bound_pre.is_explicitly_computable
actual = rc.substitute_all_modules(mlp_2_layer).get_unique(
    rc.restrict(rc.IterativeMatcher("bilinear_mlp.pre"), start_depth=5)
)
torch.testing.assert_close(inner_bound_pre.evaluate(), actual.evaluate())

# TODO: better demo of fancy push down cases + namer?

# %%

# another convenient util is drop/filter_module_inputs
dropped = drop_module_inputs(mlp_2_layer, "bilinear_mlp.input")
assert not dropped.is_explicitly_computable  # no longer computable because we don't have all symbols bound
printer.print(dropped)

# %%

# we can also extract bound symbols to a wrapping module
# note that this requires that the symbols are always bound to the same input -
# this is necessary, otherwise substitution becomes incorrect.
extracted = rc.extract_symbols_get(mlp_call, {"bilinear_mlp.w.proj_in"})
torch.testing.assert_close(extracted.evaluate(), mlp_call.evaluate())
printer.print(extracted)

# %%

# this allows for partial substitution:
printer.print(extracted.substitute())

# %%

# extracting batched inputs is a bit tricky. By default this is disallowed
with pytest.raises(rc.ExtractSymbolsBatchedInputError) as exc3:
    rc.extract_symbols_get(mlp_call, {"bilinear_mlp.input"})
print(exc3.exconly())

# %%

# but, as mentioned in the error message, we can use conform_batch_if_needed to
# handle this case (which requires conforming).

x = rc.extract_symbols_get(mlp_call, {"bilinear_mlp.input"}, conform_batch_if_needed=True)
printer.print(x)
torch.testing.assert_close(x.evaluate(), mlp_call.evaluate())

# %% [markdown]

# ### Nested batching
#
# TLDR: module batching works the way you would expect - it always adds new
# batch dims which are batched together.
#
# Modules add new batch dims as needed and can batch over each other in nested
# ways. For a complex example of this, see test_complex_batching in
# interp/circuit/interop_rust/test_module.py (I might add a single example at
# some point).
#
# This sometimes requires adding rearranges, so be aware of that.

# %% [markdown]

# ### Module rules
#
# We guarantee that substitution in any order is extentionally equal (but
# different orders may result in different rearranges and different names for
# these rearranges). Various transformation require reshaping symbols, so we
# also want validity guarantees wrt to this.
#
# First, we have to resolve the case where a parent module binds a symbol and a child
# module binds that same symbol. In this case, the inner most binding always
# takes priority. NOTE: this differs from how 'Expander' works (intentionally)!
#
# We also have to enforce the following requirements:
#
# - No higher order functions: If a module substitutes node x into the spec of a nested module, then that nested
#   module can't substitute inside of node x (this ensures we avoid higher order
#   functions, which cause all kinds of problems as well as being turing
#   complete!)
# - Spec circuits can't have 'near misses': free symbols which have equivalent
#   identification (name and uuid), but are overall not equal (typically due to
#   shape). This ensures that reshaping symbols is always fine.

# %%


# same symbol, different value - the inner value is used
s = """
'b.arg' [] Scalar 1.2
'a' Module
  'b' Module
    'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
    'b.arg' ! 'sym'
  'a.arg' [] Scalar 1.3 ! 'sym'
"""
# the outer arg is unused!
b_arg, circ_override = rc.Parser(module_check_all_inputs_used=False).parse_circuits(s)
assert circ_override.cast_module().spec.are_args_used() == [False]
assert rc.substitute_all_modules(circ_override) == b_arg
assert rc.substitute_all_modules(circ_override.cast_module().substitute()) == b_arg

# %%

# within the same module, later children are considered 'more inner' and substituted first
s = """
'b.arg' [] Scalar 1.2
'a' Module
  'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
  'a.arg' [] Scalar 1.3 ! 'sym'
  'b.arg' ! 'sym'
"""
b_arg, circ_same = rc.Parser(module_check_all_inputs_used=False).parse_circuits(s)
assert circ_same.cast_module().spec.are_args_used() == [False, True]
assert rc.substitute_all_modules(circ_same) == b_arg

# %%

# extract gets the outer most binding
orig = rc.Add(circ_override, rc.Scalar(5.0), name="new_outer")
ex = rc.extract_symbols_get(orig, "sym")
assert isinstance(ex, rc.Module)
torch.testing.assert_close(ex.evaluate(), orig.evaluate())
ex.print()

# %%

# the same applies for more/less inner children
orig = rc.Add(circ_same, rc.Scalar(5.0), name="new_outer")
ex = rc.extract_symbols_get(orig, "sym")
ex.print()
assert isinstance(ex, rc.Module)
torch.testing.assert_close(ex.evaluate(), orig.evaluate())
ex.print()

# %%

# if you push down through an override then the flattened module with have two
# args, identically to the circ_same case. Then, by default this unused arg would be
# removed (but we can disable removing it to see what that looks like).
# Removal is done using a callback. We can change this callback to a noop.
#
# See the docs for ModuleConstructCallback and ModulePusher.
double_push_down = rc.ModulePusher(flatten_modules=True, module_construct_callback=rc.ModulePusher.noop_callback())(
    circ_override, traversal=rc.new_traversal(term_early_at="sym")
)
assert circ_same == double_push_down

# %%

# higher order function case
s = """
'a' Module
  'b' Module
    'a.sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
    'same.arg' [] Scalar 1.3 ! 'b.sym' [] Symbol 33ef7e97-434e-4831-a292-d10818df4752
  'b.sym' ! 'a.sym'
"""
with pytest.raises(rc.SubstitutionCircuitHasFreeSymsBoundByNestedModuleError) as exc6:
    # by default, this would also error on 'not all args present'
    rc.Parser(module_check_all_inputs_used=False)(s)
print(exc6.exconly())

# %%

# fancier higher order function case
s = """
'a' Module
  'x' Add
    'b' Module
      'mul' Einsum ,a->a
        'a.sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
        'sub' [7] Scalar 7.3
      'same.arg' [] Scalar 1.3 ! 'b.sym' [] Symbol 33ef7e97-434e-4831-a292-d10818df4752
    'a.sym'
  'b.sym' ! 'a.sym'
"""
with pytest.raises(rc.SubstitutionCircuitHasFreeSymsBoundByNestedModuleError) as exc7:
    # by default, this would also error on 'not all args present'
    rc.Parser(module_check_all_inputs_used=False)(s)
print(exc7.exconly())

# %%

# 'near miss' case
s = """
'b' Module
  'sum' Add
    0 'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
    # note the different shape!
    1 'sym' [3] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
  'b.arg' [] Scalar 73.3874 ! 0 'sym'
"""
# the outer arg is unused!
with pytest.raises(rc.SubstitutionFoundNEQFreeSymbolWithSameIdentificationError) as exc11:
    P(s)
print(exc11.exconly())

# %% [markdown]

# ### Renaming/replace path utils
#
# it can also be useful to substitute as module with renaming

# %%

# by default we get a bunch of repeated items
printer.print(mlp_2_layer.substitute())

# but we can avoid this by using a name prefix
printer.print(mlp_2_layer.substitute(name_prefix="outer."))

# %%

# note that substituting only renames circuits which are in the replaced path
# here a more complex example so you can see what's going on
s = """
'a.norm' Module
  'ln' Add
    'ln.w.bias' [?] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487
    'ln.y_scaled' Einsum h,h->h
      'ln.y' Einsum h,->h
        'ln.mean_subbed' Add
          'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
          'ln.neg_mean' Einsum h,z,->z
            'ln.input'
            'ln.neg' [1] Scalar -1
            'ln.c.recip_hidden_size' GeneralFunction reciprocal
              'ln.c.hidden_size' GeneralFunction last_dim_size
                'ln.input'
        'ln.rsqrt' GeneralFunction rsqrt
          'ln.var_p_eps' Add
            'ln.c.eps' [] Scalar 0.00001
            'ln.var' Einsum h,h,->
              'ln.mean_subbed'
              'ln.mean_subbed'
              'ln.c.recip_hidden_size'
      'ln.w.scale' [?] Symbol 0fa341c3-34b3-4699-847f-08674808b28a
  'a.norm.input' [?,?] Symbol 6a622698-fd68-4d25-aeee-e8d38e68049e ! 'ln.input'
  'a.ln.w.bias' [?] Symbol 2c737289-2702-404c-a22e-ad37c2652620 ! 'ln.w.bias'
  'a.ln.w.scale' [?] Symbol c564149d-e226-4e3d-8e47-7d6e2ceea99e ! 'ln.w.scale'
"""
a_norm = P(s)
# (see below for extract symbols explanation)
extracted_a_norm = rc.extract_symbols_get(a_norm, {"ln.w.bias", "ln.w.scale"})
extracted_sub = extracted_a_norm.substitute(name_prefix="w.a.")
expected_extracted = """
'a.norm' Module
  'w.a.ln' Add
    'a.ln.w.bias' [?] Symbol 2c737289-2702-404c-a22e-ad37c2652620
    'w.a.ln.y_scaled' Einsum h,h->h
      'ln.y' Einsum h,->h
        'ln.mean_subbed' Add
          'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
          'ln.neg_mean' Einsum h,z,->z
            'ln.input'
            'ln.neg' [1] Scalar -1
            'ln.c.recip_hidden_size' GeneralFunction reciprocal
              'ln.c.hidden_size' GeneralFunction last_dim_size
                'ln.input'
        'ln.rsqrt' GeneralFunction rsqrt
          'ln.var_p_eps' Add
            'ln.c.eps' [] Scalar 0.00001
            'ln.var' Einsum h,h,->
              'ln.mean_subbed'
              'ln.mean_subbed'
              'ln.c.recip_hidden_size'
      'a.ln.w.scale' [?] Symbol c564149d-e226-4e3d-8e47-7d6e2ceea99e
  'a.norm.input' [?,?] Symbol 6a622698-fd68-4d25-aeee-e8d38e68049e ! 'ln.input'
"""
assert extracted_sub == P(expected_extracted)

printer.print(extracted_sub)

# %%

# we can also do map_on_replaced_path and rename on replaced path without subbing
extracted_a_norm.rename_on_replaced_path(prefix="hi").print()
extracted_a_norm.map_on_replaced_path(lambda x: x.rename(x.name.replace(".", "_"))).print()

# %%

# and this does the right thing  when the symbol is overridden by an inner module (this is just a test)
s = """
'a' Module
  'add' Add
    'b' Module
      'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
      'b.arg' [] Scalar 1.2 ! 'sym'
    'sym'
  'a.arg' [] Scalar 1.3 ! 'sym'
"""
circ_override = rc.Parser()(s).cast_module()

expected = """
'a' Module
  'hiadd' Add
    'b' Module
      'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
      'b.arg' [] Scalar 1.2 ! 'sym'
    'sym'
  'a.arg' [] Scalar 1.3 ! 'sym'
"""
assert circ_override.rename_on_replaced_path(prefix="hi") == rc.Parser()(expected)

s = """
'a' Module
  'add' Add
    'b' Module
      'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
      'add_inner' Add ! 'sym'
        'new' [] Scalar 1.7
        'sym'
    'sym'
  'a.arg' [] Scalar 1.3 ! 'sym'
"""
circ_override = rc.Parser()(s).cast_module()

expected = """
'a' Module
  'hiadd' Add
    'hib' Module
      'sym' [] Symbol c9517c8e-0b5a-4ebf-94c8-6da413c9f5d1
      'hiadd_inner' Add ! 'sym'
        'new' [] Scalar 1.7
        'sym'
    'sym'
  'a.arg' [] Scalar 1.3 ! 'sym'
"""
assert circ_override.rename_on_replaced_path(prefix="hi") == rc.Parser()(expected)

# %%

# TODO: add nice utility for renaming symbols (e.g., on conform?)

# %% [markdown]

# ### Module errors
#
# You can get really long errors due to nested substitution resulting in incompatible dtypes.
# But, the root cause should typically make it clear what went wrong!

# %%


t, _, _ = mod_l.TransformerParams(mod_l.TransformerBlockParams(), num_layers=7).garbage_call(
    hidden_size=15,
    head_size=3,
    num_heads=5,
    seq_len=11,
    batch_shape=(2, 3),
)
t = t.get_unique("t.bind_w")

# %%

print(t.device_dtype)

# %%

with pytest.raises(rc.MiscInputChildrenMultipleDtypesError) as exc8:
    rc.module_new_bind(
        t, ("t.input", rc.Array.randn(2, 3, 11, 15, device_dtype=rc.TorchDeviceDtypeOp(dtype="float64")))
    )
# this error is *huge*, but it should be pretty clear what went wrong just by the root cause
# we might improve the situation somewhat later.
print(exc8.exconly())


# %%

# random aside:
# if you call the optimizer on something which still has some symbolic sizes
# (which aren't removed by conform_all_modules), then you'll get a warning
s_with_sym = """
0 Einsum ij,jk->ik
  1 [?,?] Symbol 6594dafb-81f8-49a3-8381-6ffc4746b474
  2 [?,?] Symbol bc646cfe-a299-4580-9a54-d8d5001a566b
"""

with pytest.warns(rc.OptimizingSymbolicSizeWarning) as w:
    rc.optimize_circuit(rc.Parser()(s_with_sym))
print(w.list[0])

# %%

# another random aside: modules do various global caching. This caching needs to
# retain circuits which results in keeping those circuits around forever.
# This can result in Arrays not being freed
#
# You can clear this cache via
rc.clear_module_circuit_caches()
# to ensure that circuits are freed.
# This is somewhat of a hack.



# import os
# os.environ["RUST_BACKTRACE"] = "1"

import pytest

import rust_circuit as rc
from rust_circuit.py_utils import I

# %% [markdown]

# There are some rust circuit error handling features worth knowing about:
# - Exception per error type
# - Accessing error data
# - error traceback

# %%

softmaxed = rc.softmax(rc.Array.randn(7, 3))
idxed = softmaxed.index(I[:, 2])

# %%

# there are different exception types for most errors.
# If you go to definition on the below error, you can see the full list of exceptions and what they inherit from.
with pytest.raises(rc.PushDownIndexGeneralFunctionSomeAxesNotPossibleError) as excinfo:
    # invalid push down due to being on unbatchable dim
    rc.push_down_index(idxed)

print(excinfo.exconly())

# %%

# errors also contain additional data
try:
    rc.push_down_index(idxed)
    assert False
except rc.PushDownIndexGeneralFunctionSomeAxesNotPossibleError as e:
    assert e.args[0].index_node == idxed
    assert e.args[0].inner_rank == 2
    assert e.args[0].top_axis_indices == [2]
    assert e.args[0].num_non_batch_out == 1
    print(dir(e.args[0]))

    print()

    # you can also use:
    print(e.args[0].items())

# %% [markdown

# If you set the env var RUST_BACKTRACE=1, then errors will have a backtrace through rust
# This may or may not be helpful, but is certainly helpful when debugging issues in the rust code.
# If you rerun this nb with the top lines uncommented, you should see this.

# %% [markdown]

# ## Rust details

# On the rust side, we add additional information to errors using
# `.context(...)` which is from the anyhow library
#
# If errors ever lack context due to passing through multiple rust functions,
# this should be used.

# %%



# %%

from IPython import get_ipython

get_ipython().run_line_magic("load_ext", "autoreload")
get_ipython().run_line_magic("autoreload", "2")


import gc

from interp.circuit.circuit_model_rewrites import MLPHeadAndPosSpec, split_heads_and_positions
from rust_circuit.causal_scrubbing.dataset import Dataset

# added by tao, the notebook empirically runs faster and doesnt oom with this
gc.disable()
import os
from typing import *

import jax
import torch
from transformers import AutoTokenizer

import rust_circuit as rc
from interp.circuit.circuit_models import circuit_gpt_and_tok_embeds
from interp.circuit.projects.gpt2_gen_induction.rust_path_patching import match_nodes_except
from interp.model.model_loading import load_model
from rust_circuit.causal_scrubbing.experiment import Experiment, ExperimentEvalSettings
from rust_circuit.causal_scrubbing.hypothesis import (
    Correspondence,
    FuncSampler,
    InterpNode,
    UncondSampler,
    corr_root_matcher,
)
from rust_circuit.interop_rust import py_to_rust

jax.config.update("jax_platform_name", "cpu")

# %% [markdown]

# This notebook is intended as a basic introduction to the causal scrubbing API. It is operating on the circuit that
# predicts incrementing numbers once it's seen a sequence of numbers. This experiment doesn't use the full functionality
# of causal scrubbing - it just checks whether these nodes are used in the circuit for predicting next numbers.
#
# Before running, make sure you're using the latest rust_circuit version - run `pip install rust_circuit --upgrade`
#
# This is still a work in progress. Here are some things I want to do to improve it:
# - [ ] New model loader once splitting exists for that model loader
# - [ ] Use a real hypothesis that actually talks about what the circuit does rather than just "is it important"
# - [ ] This circuit isn't exhaustive - its depth is only 2. Keep expanding it

# %% [markdown]

# First, load a tokenizer and the data we'll use for this experiment
#
# I load a token for every number because, later on, we'll be using this to determine incorrect results
#
# The sequences are just '<|endoftext|> 4 5 6' (for example) with random starting points between 1 and 100.
#
# We'll be scrubbing out unimportant nodes with other number sequences so we can isolate the nodes that are involved
# in incrementing the numbers directly - not in checking if the sequence is incrementing, for example, or in checking
# whether the tokens are numbers (even though these heuristics are likely useful for implementing the overall behavior).

# %%

tokenizer = AutoTokenizer.from_pretrained("gpt2")
tokenizer.pad_token = tokenizer.eos_token

all_numbers = [f" {i}" for i in range(1, 104)]
all_number_tokens = tokenizer(all_numbers, padding=True, return_tensors="pt").input_ids.squeeze(-1)

start_points = list(range(1, 101))
sequences = [
    tokenizer.bos_token + "".join(f" {i}" for i in range(start_point, start_point + 3)) for start_point in start_points
]

tokens = tokenizer(sequences, padding=True, return_tensors="pt").input_ids

correct_values = "".join(f" {start_point + 3}" for start_point in start_points)

correct_tokens = tokenizer(correct_values, padding=True, return_tensors="pt").input_ids.squeeze(0)

# %% [markdown]

# Some utilities we'll be using later on

# %%


def logit_diff(logits: torch.Tensor, labels: rc.Array) -> float:
    """
    Returns difference between probability of correct number and all numbers [1, 103]. Will always be negative, since
    the correct number is one of those numbers. Higher values (closer to zero) are more correct.
    """
    correct_probs = logits[:, 3].gather(-1, labels.value.cpu().long().unsqueeze(-1)).logsumexp(dim=-1)
    all_number_probs = logits[:, 3].gather(-1, all_number_tokens.unsqueeze(0)).logsumexp(dim=-1)
    return (correct_probs - all_number_probs).mean().item()


def loss_fn(logits: torch.Tensor, labels: rc.Array) -> float:
    return torch.nn.functional.cross_entropy(logits[:, 3], labels.value.cpu().long()).item()


def print_top_k(tokenizer: AutoTokenizer, model_res: torch.Tensor):
    """
    Useful for debugging what the model has chosen to output
    """
    print(model_res[-1:].softmax(-1).topk(10))
    print(tokenizer.batch_decode(model_res[-1:].topk(10).indices.T))  # type: ignore


def node_loc_to_name(
    layer: int,
    head: Union[int, Literal["mlp"]],
    position: int,
) -> str:
    """
    Find node name from location info
    """
    if head == "mlp":
        return f"m{layer}_t{position}"
    else:
        return f"a{layer}.out_h{head}_t{position}"


SameLabelSampler = lambda: FuncSampler(lambda ds: ds.labels.value.long())


# %% [markdown]

# # Load GPT model
#
# For historical reasons, we load the GPT-2 model from Jax before loading it into circuits
#
# TODO: rewrite when new loaders exist

# %%

# load GPT model
models_dir_local = os.path.expanduser("~/rrfs/interpretability_models_jax")

gpt2, gpt2_vars, _ = load_model("gpt2", models_dir=models_dir_local)
gpt2_b = gpt2.bind(gpt2_vars)


# %% [markdown]

# # Next, convert the model to a circuit
#
# First, we need to call a model with `circuit_gpt_and_tok_embeds` to convert it to a full circuit. This will create a
# full computation graph for the model call with this input.
#
# Since we're going to be scrubbing the inputs later, we can put in arbitrary data for now. It just needs to be the
# same shape as the data you'll scrub in later. There's also no need to create different circuits for models called on
# different inputs.

# %%

# convert model to circuit

c_logprobs, tok_inp = circuit_gpt_and_tok_embeds(
    gpt2_b,
    torch.ones_like(tokens[:1]),  # arbitrary data
    use_batch=True,
    output_type="log_probs",
)


# %% [markdown]

# # Split the model by head
#
# Because attention heads are implemented as one matrix multiplication in the model, by default they don't have
# different nodes in the computation graph. Similarly,
#
# The split heads will have names that look like `m{layer}_at_h{head}` for MLPs or `a{layer}_at_h{head}_at_t{position}`
# for heads.

# %%

# split model by head
nodes_to_split = [
    MLPHeadAndPosSpec(layer, head_or_mlp, pos)  # type: ignore
    for layer in range(12)
    for head_or_mlp in list(range(12)) + ["mlp"]
    for pos in range(1, 4)
]

split_circuit = split_heads_and_positions(c_logprobs, nodes_to_split)

# %% [markdown]

# Next, move the model to rust circuits. TODO: remove this once we move splitting to rust

# %%

circuit = py_to_rust(split_circuit, rc.TorchDeviceDtypeOp("cuda:0", "float32"))


# %%

# This is an alternate way to load the model, pre-split for us. It also replaces the inputs with DiscreteVars. The code
# below expects this model, so you should run this cell or things won't work.

with open(os.path.expanduser("~/rrfs/adria/gpt2small/seq_len_4.txt")) as f:
    circuit = rc.Parser()(f.read())

# %% [markdown]

# # Next, make a Dataset
#
# A vanilla Dataset takes (tensor, tensor) (or numpy array). It can sample from and shuffle that dataset.
#
# The commented-out version is used for replacing the tok_embeds node instead, because GPT-2 doesn't have a tokens node
# by default (but adria's version does)

# %%

dataset = Dataset(
    (
        rc.Array(tokens.float().to("cuda:0"), "tokens"),
        rc.Array(correct_tokens.squeeze(-1).float().to("cuda:0"), "labels"),
    ),
    input_names={"tokens"},
)

# Alternate dataset for handmade model

# dataset = Dataset(
#     np.array(gpt2_b.embedding.token_embedding(incr_dataset.tokens.long().numpy())),
#     incr_dataset.correct_tokens.squeeze(-1).long(),
# )

# %% [markdown]

# Let's test a simple hypothesis: everything in this graph matters. To do this, we're going to start by creating several
# objects:
#
# - Correspondence maps from nodes in the model to an interpretation of the model
# - ModelBranch is a single edge from one node to another that we think is relevant for the model (plus there's an extra
#   one for the root node)
# - InterpNode represents a leaf node. The name is only for printing, so I'm using the same names as the circuit
#   nodes themselves. The important feature here is the cond_sampler, which is used to choose how to sample the
#   activations for the model node associated with this InterpNode. In this case, we enforce that the activations must
#   be the same as another input with the same label (SameLabelSampler).
# - Experiment takes a Circuit, a Dataset, and a Correspondance, and automatically re-samples according to the causal
#   scrubbing algorithm.
#
# You create matching ModelBranches and InterpNodes (making sure they descend from corresponding MB/INs)
#
# You must start with a root node. This will match the root of the circuit.
#
# Notice that we don't hook any other branches/nodes up to this node in this case. This is because, unless you
# explicitly connect every node to the input node (in this case "tokens"), Experiment.run will automatically include
# every path from every leaf node to the inputs in your hypothesis. In this case, this will include every path from the
# root node to the inputs, so it includes the entire circuit. This is just saying that the entire circuit should be
# included.
#
# First, let's print the circuit.
# %%

# everything matters hypothesis

corr = Correspondence()

m_log_probs = corr_root_matcher
log_probs = InterpNode(name="log_probs", cond_sampler=SameLabelSampler())
corr.add(log_probs, m_log_probs)

everything_experiment = Experiment(
    circuit,  # type: ignore
    dataset,  # type: ignore
    corr,
)
everything_scrubbed = everything_experiment.scrub(100)
everything_scrubbed.print()

# %%
eval_settings = ExperimentEvalSettings(optimize=False)
everything_output = everything_scrubbed.evaluate(eval_settings)

# TODO: explain why we're using the labels from CS
everything_logit_diff = logit_diff(everything_output.cpu(), everything_scrubbed.ref_ds.labels)
print("logit diff (less negative is better):", everything_logit_diff)

everything_loss = loss_fn(everything_output.cpu(), everything_scrubbed.ref_ds.labels)
print("loss (lower is better):", everything_loss)

# %% [markdown]

# Now let's do the same thing, but with `cond_sampler=UncondSampler()`. This will make the sampler sample randomly from
# all inputs regardless of the label, so it'll scramble all our activations.

# Our goal will be to recover something close to the above value, starting at this value.

# %%

# nothing matters hypothesis

corr = Correspondence()  # maybe different input name?

m_log_probs = corr_root_matcher
logits = InterpNode(name="log_probs", cond_sampler=UncondSampler())
corr.add(logits, m_log_probs)

uncond_experiment = Experiment(
    circuit,  # type: ignore
    dataset,  # type: ignore
    corr,
)

uncond_scrubbed = uncond_experiment.scrub(100)
uncond_scrubbed.print()

# %%

uncond_outputs = uncond_scrubbed.evaluate(eval_settings)
uncond_logit_diff = logit_diff(uncond_outputs.cpu(), uncond_scrubbed.ref_ds.labels)
print("logit diff (less negative is better):", uncond_logit_diff)

uncond_loss = loss_fn(uncond_outputs.cpu(), uncond_scrubbed.ref_ds.labels)
print("loss (lower is better):", uncond_loss)

# When recovering loss do we use


# %% [markdown]

# For pedagogical purposes, let's compare this to running the model with a set of inputs. You'll see that the results
# are similar but not the same, because the input values are different. CS will sample from our input tokens, while this
# code will use exactly our selected input tokens.

# To swap out the inputs, we need to replace the input DiscreteVars that are there by default with our own.

# Why do we need to do this? Because DiscreteVars represent distributions over values, not a specific value.

# We can replace the original DiscreteVar with one that uses the values we care about, and then we can sample from that
# distribution to get a Circuit we can evaluate.

# %%

# compare to original model

# new input var

input_var = rc.DiscreteVar.new_uniform(rc.Array(tokens.to("cuda:0").float()))

# replace tokens

orig_toks_circuit = rc.Updater(lambda c: input_var)(circuit, "tokens")

# sample using a spec that samples every value once

orig_results = rc.Sampler(rc.RunDiscreteVarAllSpec([input_var.probs_and_group])).sample(orig_toks_circuit).evaluate()

orig_logit_diff = logit_diff(orig_results.cpu(), rc.Array(correct_tokens))
print("logit diff (less negative is better):", orig_logit_diff)

orig_loss = loss_fn(orig_results.cpu(), rc.Array(correct_tokens))
print("loss (lower is better):", orig_loss)

# %% [markdown]

# Now let's try a real graph that implements incrementing numbers. We will add branches and nodes to the correspondence
# to make sure we represent our fullest hypothesis. Again, any leaf nodes include all paths to the tokens, so this
# hypothesis is actually much more generous to us than it should be.
#
# So you don't have to run a lot of cells, I've just added comments inline to walk through what we're doing.
#
# One important new feature: When adding a new branch off an existing one, if you use `to`, it'll go through all nodes
# in the circuit by default. This will make your hypothesis a lot less specific. If you don't want this, yo uneed to use
# match_nodes_except as we do here.
#
# Note that the printed experiment is much more interesting now - the relevant nodes we've selected are highlighted.
#
# Also note that we recover most of the loss!


# %%

# small hypothesis

# List heads/mlps we think are involved

heads: list[MLPHeadAndPosSpec] = [
    MLPHeadAndPosSpec(10, 2, 3),
    MLPHeadAndPosSpec(9, 1, 3),
    MLPHeadAndPosSpec(8, 11, 3),
    MLPHeadAndPosSpec(7, 10, 3),
]
mlps: list[MLPHeadAndPosSpec] = [
    MLPHeadAndPosSpec(10, "mlp", 3),
    MLPHeadAndPosSpec(9, "mlp", 3),
    MLPHeadAndPosSpec(8, "mlp", 3),
]

# Create basic correspondence

corr = Correspondence()

m_log_probs = corr_root_matcher
log_probs = InterpNode(name="log_probs", cond_sampler=SameLabelSampler())
corr.add(log_probs, m_log_probs)

# add log_probs -> heads/MLPs

for layer, head, position in [*heads, *mlps]:
    name = node_loc_to_name(layer, head, position)
    m_head = rc.restrict(
        m_log_probs.chain(name), term_early_at=match_nodes_except([node_loc_to_name(layer, head, position)])
    )
    i_head = log_probs.make_descendant(name=name, cond_sampler=SameLabelSampler())
    corr.add(i_head, m_head)

# create experiment

small_experiment = Experiment(
    circuit,  # type: ignore
    dataset,  # type: ignore
    corr,
)

small_scrubbed = small_experiment.scrub(100)
small_scrubbed.print()

# %%
small_outputs = small_scrubbed.evaluate(eval_settings)

# TODO: explain why we're using the labels from CS
small_logit_diff = logit_diff(small_outputs.cpu(), small_scrubbed.ref_ds.labels)
print("logit diff (less negative is better):", small_logit_diff)

small_loss = loss_fn(small_outputs.cpu(), small_scrubbed.ref_ds.labels)
print("loss (lower is better):", small_loss)

# %% [markdown]

# Now let's calculate the loss recovered so far.

# %%

print((small_loss - uncond_loss) / (everything_loss - uncond_loss))

# %% [markdown]

# This is a more complete graph. It won't recover as much loss, unfortunately, because it's actually checking a smaller
# subset of the model because it doesn't include all paths from the relevant heads and MLPs to the origin. There's often
# a tradeoff between specficity and accuracy.

# %%

# real hypothesis

# More heads we know about

inhibition: list[MLPHeadAndPosSpec] = [MLPHeadAndPosSpec(10, 7, 3)]

# Create basic correspondence

corr = Correspondence()

m_log_probs = corr_root_matcher
log_probs = InterpNode(name="log_probs", cond_sampler=SameLabelSampler())
corr.add(log_probs, m_log_probs)

# add log_probs -> heads (which tend not to look at other heads)

for layer, head, position in heads:
    name = node_loc_to_name(layer, head, position)
    m_head = rc.restrict(
        m_log_probs.chain(name), term_early_at=match_nodes_except([node_loc_to_name(layer, head, position)])
    )
    i_head = log_probs.make_descendant(name=name, cond_sampler=SameLabelSampler())
    corr.add(i_head, m_head)

# add log_probs -> mlps -> other heads and mlps

for layer, _, position in mlps:
    name = node_loc_to_name(layer, "mlp", position)
    m_mlp = rc.restrict(
        m_log_probs.chain(name), term_early_at=match_nodes_except([node_loc_to_name(layer, "mlp", position)])
    )
    i_mlp = log_probs.make_descendant(name=name, cond_sampler=SameLabelSampler())
    corr.add(i_mlp, m_mlp)

    for h_layer, h_head, h_position in heads + mlps:
        if h_layer > layer or h_layer == layer and h_head == "mlp":
            continue
        h_name = node_loc_to_name(h_layer, h_head, h_position)
        m_to_head = rc.restrict(
            m_mlp.chain(h_name),
            term_early_at=match_nodes_except(
                [node_loc_to_name(layer, "mlp", position), node_loc_to_name(h_layer, h_head, h_position)]
            ),
        )
        i_to_head = i_mlp.make_descendant(name=h_name, cond_sampler=SameLabelSampler())
        corr.add(i_to_head, m_to_head)

# add inhibition heads

for layer, head, position in inhibition:
    name = node_loc_to_name(layer, head, position)
    m_mlp = rc.restrict(
        m_log_probs.chain(name), term_early_at=match_nodes_except([node_loc_to_name(layer, head, position)])
    )
    i_mlp = log_probs.make_descendant(name=name, cond_sampler=SameLabelSampler())
    corr.add(i_mlp, m_mlp)

tree_experiment = Experiment(
    circuit,  # type: ignore
    dataset,  # type: ignore
    corr,
)
tree_scrubbed = tree_experiment.scrub(100)
# tree_scrubbed.print()

# %%

tree_outputs = tree_scrubbed.evaluate(eval_settings)

tree_logit_diff = logit_diff(tree_outputs.cpu(), tree_scrubbed.ref_ds.labels)
print("logit diff (less negative is better):", tree_logit_diff)

tree_loss = loss_fn(tree_outputs.cpu(), tree_scrubbed.ref_ds.labels)
print("loss (lower is better):", tree_loss)

# %% [markdown]

# Here's the percentage of loss recovered - 89%, still pretty good!

# %%

print((tree_loss - uncond_loss) / (everything_loss - uncond_loss))

# %%



# this file is included with `include_str!()` in py_types.rs
import functools
import itertools
import math
from typing import Any, List, Optional, Sequence, Tuple, Union, cast

import torch
import torch.nn.functional


def check_type(x, type_v):
    if not isinstance(x, type_v):
        raise TypeError(f"expected type {repr(type_v)} but got {type(x)}")


def get_tensor_shape(x):
    if not isinstance(x, torch.Tensor):
        raise TypeError(f"expected type {repr(torch.Tensor)} but got {type(x)}")
    return x.shape, x.device, x.dtype


dtype_from_string = {
    "float32": torch.float32,
    "float64": torch.float64,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
    "int64": torch.int64,
    "int32": torch.int32,
    "int16": torch.int16,
    "int8": torch.int8,
    "bool": torch.bool,
    "uint8": torch.uint8,
}
dtype_to_string = {v: k for k, v in dtype_from_string.items()}


def maybe_dtype_to_maybe_string(x):
    return dtype_to_string.get(x)


def scalar_to_tensor(scalar, shape, device_dtype):
    "Makes a stride 0 repeat view into a 1 element tensor"
    scalar_tensor = torch.tensor(scalar, device=device_dtype.device, dtype=dtype_from_string[device_dtype.dtype])
    return torch.broadcast_to(scalar_tensor, tuple(shape))


def tensor_scale(tensor):
    return tensor.abs().mean().cpu().item()


def cast_tensor(tensor, device_dtype):
    kwargs = {}
    if device_dtype.device is not None:
        kwargs["device"] = device_dtype.device
    if device_dtype.dtype is not None:
        kwargs["dtype"] = dtype_from_string[device_dtype.dtype]
    return tensor.to(**kwargs)


zero_tensor = torch.zeros(())


def un_flat_concat(tensor: torch.Tensor, shapes) -> List[torch.Tensor]:
    lens = [math.prod(x) for x in shapes]
    flats = torch.split(tensor, lens, dim=0)
    return [x.reshape(shape) for x, shape in zip(flats, shapes)]


def log_exp_p_1_fn(x: torch.Tensor):
    # piecewise to nicely handle numerics
    addr = 1.0
    return torch.where(x < 0.0, torch.log(torch.exp(x) + addr), torch.log(1.0 + torch.exp(-x) * addr) + x)


generalfunctions = {
    "sin": torch.sin,
    "cos": torch.cos,
    "gelu": torch.nn.functional.gelu,
    "gelu_new": lambda x: 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0)))),
    "relu": torch.relu,
    "step": lambda x: torch.where(
        x > 0.0, torch.ones((), dtype=x.dtype, device=x.device), torch.zeros((), dtype=x.dtype, device=x.device)
    ),
    "sigmoid": torch.sigmoid,
    "rsqrt": torch.rsqrt,
    "reciprocal": torch.reciprocal,
    "tanh": torch.tanh,
    "softmax": functools.partial(torch.softmax, dim=-1),
    "log_softmax": functools.partial(torch.log_softmax, dim=-1),
    "log_exp_p_1": log_exp_p_1_fn,
    "gaussian_pdf": lambda x: torch.exp(-(x ** 2) / 2) / math.sqrt(2 * math.pi),
    "gaussian_cdf": lambda x: torch.distributions.normal.Normal(0, 1).cdf(x),
    "q_from_qr": lambda x: torch.linalg.qr(x)[0],
    "min": lambda x: torch.min(x, dim=-1)[0],
    "max": lambda x: torch.max(x, dim=-1)[0],
    "last_dim_size": lambda x: torch.full(x.shape[:-1], x.shape[-1], dtype=x.dtype, device=x.device),
    "abs": torch.abs,
    "exp": torch.exp,
    "log": torch.log,
    "logit": torch.logit,
}

pow = torch.pow


def check_canon_idx(i: int, count: int):
    assert count >= 0, count
    if i >= 0:
        assert i < count, (i, count)
    else:
        assert i >= -count, (i, count)
    return i % count


def check_ints(x: torch.Tensor):
    assert (x.long().to(dtype=x.dtype) == x).all()


def gen_index_function(
    x: torch.Tensor, index: torch.Tensor, index_dim: int, batch_x: bool, batch_index: bool, check_index_ints: bool
):
    # copy of  WildIndex from computational_node.py
    if check_index_ints:
        check_ints(index)
    use_index_as_prefix = batch_x and batch_index
    if use_index_as_prefix:
        return explicit_gen_index_function(x, index, index_dim, x.ndim - index.ndim, check_index_ints=check_index_ints)
    elif batch_index:
        moved_out = torch.index_select(x, index_dim, index.long().flatten()).moveaxis(index_dim, 0)
        return moved_out.reshape((*index.shape, *moved_out.shape[1:]))
    else:
        canon_index_dim = check_canon_idx(index_dim, x.ndim)
        out = torch.index_select(x, canon_index_dim, index.long().flatten())
        return out.reshape((*out.shape[:canon_index_dim], *index.shape, *out.shape[canon_index_dim + 1 :]))


def explicit_gen_index_function(
    x: torch.Tensor, index: torch.Tensor, index_dim: int, x_non_batch_dims: int, check_index_ints: bool
):
    if check_index_ints:
        check_ints(index)

    batch_len = x.ndim - x_non_batch_dims
    batch_shape = x.shape[:batch_len]
    assert batch_shape == index.shape[:batch_len]
    # x_len = x.ndim - prefix_len
    moved_x = x.movedim(check_canon_idx(index_dim, x_non_batch_dims) + batch_len, 0)
    if moved_x.ndim == 1:
        return moved_x[index.long()]

    # end_dim is inclusive, so this is right
    flattened_x = moved_x.flatten(start_dim=1, end_dim=batch_len)
    flattened_index = index.flatten(start_dim=0, end_dim=batch_len - 1).long()

    together_arange = torch.arange(math.prod(batch_shape))
    pad_arange = together_arange[(slice(None), *([None] * (index.ndim - batch_len)))]
    new_x = flattened_x[flattened_index, pad_arange]
    return new_x.reshape((*index.shape, *moved_x.shape[1 + batch_len :]))


assert_tensors_close = torch.testing.assert_close


def make_diagonal(tensor: torch.Tensor, tensor_ints: Tuple[int, ...], out_ints: Tuple[int, ...]):
    int_sizes = {x: tensor.shape[i] for i, x in enumerate(tensor_ints)}
    deduped_shape = [int_sizes[x] for x in tensor_ints]
    result = torch.zeros(tuple(int_sizes[x] for x in out_ints), dtype=tensor.dtype, device=tensor.device)
    normal_strides_out = result.stride()
    fancy_strides = []
    for i in tensor_ints:
        indices = [j for j, x in enumerate(out_ints) if x == i]
        strides = [normal_strides_out[k] for k in indices]
        stride_here = sum(strides)
        fancy_strides.append(stride_here)
    fancy_strided = torch.as_strided(result, deduped_shape, fancy_strides)
    fancy_strided += tensor
    return result


def tensor_to_bytes(x: torch.Tensor):
    return x.cpu().numpy().tobytes()


def tensor_from_bytes(device_dtype, shape, bytes, count):
    return (
        torch.frombuffer(bytes, count=count, offset=0, dtype=dtype_from_string[device_dtype.dtype])
        .reshape(shape)
        .to(device=device_dtype.device)
    )


einsum = lambda tensors_and_axes, out_axes: torch.einsum(*itertools.chain(*tensors_and_axes), out_axes)


def random_indices(
    probs: torch.Tensor, shape: Sequence[int], replacement: bool, seed: Optional[Union[int, torch.Tensor]] = None
) -> torch.Tensor:
    shape = tuple(shape)

    assert probs.ndim >= 1

    generator: Optional[torch.Generator] = (
        torch.Generator(device=probs.device).manual_seed(int(seed)) if seed is not None else None
    )

    return torch.multinomial(
        probs.reshape(math.prod(probs.shape[:-1]), probs.shape[-1]),
        num_samples=math.prod(shape),
        replacement=replacement,
        generator=generator,
    ).reshape((*probs.shape[:-1], *shape))


# we use different axis layout than pytorch,
# batch_dims... height_width_ect... channels
# we also allow asymmetric padding to support InceptionV1
# which we have to do manually bc torch.conv doesn't support
def conv(dim, input, filter, stride, padding):
    import einops

    assert len(filter.shape) == 2 + dim, "don't support filter batching yet"
    stride = tuple(stride)

    batch_rank = len(input.shape) - dim - 1
    char_at = lambda i: chr(ord("a") + i)
    input_rearrange_string_i = " ".join([char_at(x) for x in range(len(input.shape))])
    input_rearrange_string_o = f"({' '.join([char_at(x) for x in range(batch_rank)])}) {char_at(len(input.shape)-1)} {' '.join([char_at(x) for x in range(batch_rank,len(input.shape)-1)])}"
    input_rearrange_string = f"{input_rearrange_string_i} -> {input_rearrange_string_o}"
    input_batches_together = einops.rearrange(input, input_rearrange_string)
    if any([x[0] != x[1] for x in padding]):
        padding_min = tuple([min(x) for x in padding])
        input_batches_together = torch.nn.functional.pad(
            input_batches_together,
            tuple(itertools.chain(*[(p[0] - pmin, p[1] - pmin) for p, pmin in zip(padding, padding_min)])),
        )
    padding = tuple([min(x) for x in padding])
    if dim == 1:
        filter = einops.rearrange(filter, "o a i -> o i a")
        result = torch.nn.functional.conv1d(input_batches_together, filter, None, stride, padding)
    elif dim == 2:
        filter = einops.rearrange(filter, "o a b i -> o i a b")
        result = torch.nn.functional.conv2d(input=input_batches_together, weight=filter, stride=stride, padding=padding)
    elif dim == 3:
        filter = einops.rearrange(filter, "o a b c i -> o i a b c")
        result = torch.nn.functional.conv3d(input=input_batches_together, weight=filter, stride=stride, padding=padding)
    else:
        raise ValueError("conv only supports 1, 2, or 3d convolutions")
    return einops.rearrange(
        result,
        f"{input_rearrange_string_o} -> {input_rearrange_string_i}",
        **{char_at(i): input.shape[i] for i in range(batch_rank)},
    )


class OptimizingSymbolicSizeWarning(UserWarning):
    ...


class OptimizingModulesWarning(UserWarning):
    ...


def random_i64():
    return int(torch.randint(-(2 ** 63), 2 ** 63 - 1, size=(), dtype=torch.long))


def canon_index_tensor(x: torch.Tensor) -> Union[slice, torch.Tensor]:
    if len(x) == 0:
        return slice(0, 0)
    if torch.all(torch.diff(x) == 1):
        return slice(int(x[0]), int(x[0]) + len(x))
    return x


def index_union_rebase(indices, l: int, device: str):
    mask = torch.zeros(l, dtype=torch.bool, device=device)
    for idx in indices:
        mask[idx] = True
    nonzero = torch.nonzero(mask, as_tuple=True)[0]
    rebased: List[Any] = []
    if any([isinstance(x, torch.Tensor) for x in indices]):
        place_in_new = torch.zeros(l, dtype=torch.int64, device=device)
        place_in_new[nonzero] = torch.arange(len(nonzero), dtype=torch.int64, device=device)
    for idx in indices:
        if isinstance(idx, int):
            rebased.append(torch.nonzero(nonzero == idx, as_tuple=True)[0].item())
        elif isinstance(idx, slice):
            # we know everything in the slice is in nonzero, so we can find start and take then next delta elements
            start_idx = torch.nonzero(nonzero == idx.start, as_tuple=True)[0].item()
            rebased.append(slice(start_idx, start_idx + (idx.stop - idx.start)))
        else:
            rebased.append(canon_index_tensor(place_in_new[idx]))
    # print(f"index_union_rebase in: {indices} l: {l} out: {nonzero} {rebased}")
    return (canon_index_tensor(nonzero), rebased)


ISLICE = slice(None, None)
TorchAxisIndex = Union[int, slice, torch.Tensor]
TorchIndex = Union[Tuple[TorchAxisIndex, ...], TorchAxisIndex]


def index_dimwise_to_numpy_style(index):
    result: List[List[TorchAxisIndex]] = []
    for i, idx in enumerate(index):
        if len(result) == 0 or (
            isinstance(idx, torch.Tensor) and len([x for x in result[-1] if isinstance(x, torch.Tensor)]) > 0
        ):
            result.append([ISLICE for z in index[:i] if not isinstance(z, int)])
        result[-1].append(idx)
    assert (
        len(result) == len([x for x in index if isinstance(x, torch.Tensor)])
        or len([x for x in index if isinstance(x, torch.Tensor)]) == 0
        and len(result) == 1
    ), (index, result)

    return [tuple(x) for x in result]


def index_apply_dimwise(tensor: torch.Tensor, index: TorchIndex):
    if not isinstance(index, tuple):
        index = (index,)
    result = tensor
    indices = index_dimwise_to_numpy_style(index)
    for idx in indices:
        result = result[idx]
    return result


def index_axis_to_tensor(x: TorchAxisIndex, l: int, device):
    return torch.arange(l, device=device)[x]


def shape_to_strides(shape: Tuple[int, ...]) -> Tuple[int, ...]:
    if shape == ():
        return ()
    return functools.reduce(lambda a, n: (a[0] * n,) + a, reversed(shape[1:]), cast(Tuple[int, ...], (1,)))


def join_indices_unsync(indices: Tuple[TorchAxisIndex, ...], shape: Tuple[int, ...], device):
    indices_tensors = [index_axis_to_tensor(x, l, device=device) for x, l in zip(indices, shape)]
    strides = shape_to_strides(shape)
    rank = len(indices)
    broadcasted: List[torch.Tensor] = [
        x[(() if len(x.shape) == 0 else (ISLICE,)) + ((None,) * (rank - (i + 1)))] * stride
        for i, (x, stride) in enumerate(zip(indices_tensors, strides))
    ]
    summed = functools.reduce(lambda a, b: a + b, broadcasted[1:], broadcasted[0])
    result = summed.flatten()
    if all(isinstance(x, int) for x in indices):
        return result.item()
    return canon_index_tensor(result)


def just_join_indices_sync(indices: Tuple[TorchAxisIndex, ...], shape: Tuple[int, ...], device) -> torch.Tensor:
    indices_tensors = [index_axis_to_tensor(x, l, device) for x, l in zip(indices, shape)]
    strides = shape_to_strides(shape)
    indices_raised: List[torch.Tensor] = [x * s for x, s in zip(indices_tensors, strides)]
    return functools.reduce(lambda a, b: a + b, indices_raised[1:], indices_raised[0])


def scatter(idx, shape, tensor):
    result = torch.zeros(shape, dtype=tensor.dtype, device=tensor.device)
    if len([x for x in idx if isinstance(x, torch.Tensor)]) <= 1:
        result[idx] = tensor
    else:
        raise NotImplementedError(
            "scatter where index has multiple tensor axes that need to be applied dimwise unsupported"
        )
    return result



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Tuple

from dataclasses import dataclass

import torch

import rust_circuit.index_util.internal.resolve as resolve

#
# Planning
#


@dataclass
class SizeVar:
    arr_number: int
    axis: int


@dataclass
class SizeEqConstraint:
    lhs: SizeVar
    rhs: SizeVar


@dataclass
class ShapesPlan:
    arr_ranks: list[int]
    axis_sizes: list[Optional[SizeVar]]
    constraints: list[SizeEqConstraint]


def plan_shapes(num_axes: int, array_indices: list[list[resolve.IdxExpr]]) -> ShapesPlan:
    arr_ranks = [len(indices) for indices in array_indices]

    axis_sizes: list[Optional[SizeVar]] = [None] * num_axes
    constraints: list[SizeEqConstraint] = []

    for arr_number, indices in enumerate(array_indices):
        for i, idx in enumerate(indices):
            if not isinstance(idx, resolve.OutAxisIdx):
                continue
            axis = idx.axis
            prev = axis_sizes[axis]
            if prev is None:
                axis_sizes[axis] = SizeVar(arr_number, i)
            else:
                constraints.append(SizeEqConstraint(prev, SizeVar(arr_number, i)))

    return ShapesPlan(arr_ranks=arr_ranks, axis_sizes=axis_sizes, constraints=constraints)


#
# Execution
#


@dataclass
class RuntimeShapes:
    batch_shape: list[int]
    axis_sizes: list[int]


def execute_shapes_plan(
    plan: ShapesPlan, arrays: list[Tuple[str, list[int]]], user_axis_sizes: Optional[list[Optional[int]]] = None
) -> RuntimeShapes:
    assert len(arrays) == len(plan.arr_ranks)

    opt_axis_sizes: list[Optional[int]]
    if user_axis_sizes is None:
        opt_axis_sizes = [None] * len(plan.axis_sizes)
    else:
        opt_axis_sizes = user_axis_sizes

    if len(opt_axis_sizes) < len(plan.axis_sizes):
        raise ValueError(f"Expected at least {len(plan.axis_sizes)} axis sizes, got {len(opt_axis_sizes)}")

    axis_sizes_batch_rank = len(opt_axis_sizes) - len(plan.axis_sizes)

    batch_shapes: list[list[int]] = []
    batch_shapes.append([size if size is not None else 1 for size in opt_axis_sizes[:axis_sizes_batch_rank]])
    for arr_rank, (name, shape) in zip(plan.arr_ranks, arrays):
        if len(shape) < arr_rank:
            raise ValueError(f"Expected array {name!r} to have rank at least {arr_rank}; got shape {shape}")
        batch_shapes.append(shape[: len(shape) - arr_rank])

    try:
        batch_shape = list(torch.broadcast_shapes(*batch_shapes))
        assert all(x == batch_shape for x in batch_shapes[1:])
    except RuntimeError as e:
        raise ValueError(f"Batch dimensions are incompatible")

    def get_size_var(size_var: SizeVar) -> int:
        arr_shape = arrays[size_var.arr_number][1]
        arr_batch_rank = len(arr_shape) - plan.arr_ranks[size_var.arr_number]
        return arr_shape[arr_batch_rank + size_var.axis]

    axis_sizes: list[int] = []
    for i, size_var in enumerate(plan.axis_sizes):
        if size_var is None:
            size = opt_axis_sizes[axis_sizes_batch_rank + i]
            if size is None:
                raise ValueError(f"Could not infer size for axis {i - len(plan.axis_sizes)}")
            axis_sizes.append(size)
        else:
            axis_sizes.append(get_size_var(size_var))

    for constraint in plan.constraints:
        lhs_size = get_size_var(constraint.lhs)
        rhs_size = get_size_var(constraint.rhs)
        if lhs_size != rhs_size:
            lhs_num = constraint.lhs.arr_number
            rhs_num = constraint.rhs.arr_number
            lhs_name = arrays[lhs_num][0]
            rhs_name = arrays[rhs_num][0]
            lhs_rank = plan.arr_ranks[lhs_num]
            rhs_rank = plan.arr_ranks[rhs_num]
            lhs_expr = f"{lhs_name}.shape[{constraint.lhs.axis - lhs_rank}]"
            rhs_expr = f"{rhs_name}.shape[{constraint.rhs.axis - rhs_rank}]"
            raise ValueError(
                f"Incompatible shapes: "
                + f"expected {lhs_expr} == {rhs_expr}; "
                + f"got {lhs_expr} = {lhs_size}, {rhs_expr} = {rhs_size}"
            )

    return RuntimeShapes(batch_shape=batch_shape, axis_sizes=axis_sizes)



from __future__ import annotations

import asyncio
import json
import socket
import threading
import urllib.parse
from functools import partial
from threading import Lock
from typing import *
from typing import Any

import websockets
from IPython.core.display import HTML, display

from rust_circuit.ui.circuits_very_named_tensor import CircuitsVeryNamedTensor as rCircuitsVeryNamedTensor
from rust_circuit.ui.encoding import get_callback_result, msgpack_encode
from rust_circuit.ui.very_named_tensor import LazyVeryNamedTensor, VeryNamedTensor

served_things: Dict[str, Any] = {}
thread = None
lock = Lock()
sockets: Dict[str, socket.socket] = {}
global_frontend_url = None
global_backend_url = None
global_port = None
debug = False


async def handler(websocket):
    global sockets
    if debug:
        print("handler")
    name = None
    try:
        async for buf in websocket:
            msg = json.loads(buf)
            kind = msg.get("kind")
            if debug:
                print(kind)
            if kind == "init":
                pass
            elif kind == "nameStartup":
                name = msg["name"]
                if debug:
                    print("name", name)
                sockets[name] = websocket
                with lock:
                    thing = served_things.get(name, None)
                await websocket.send(msgpack_encode({"kind": "nameStartup", "data": thing}))
            elif kind == "callback":
                enc = msgpack_encode(get_callback_result(msg))
                await websocket.send(enc)
            else:
                print("Unrecognized message: ", msg)
    except websockets.exceptions.ConnectionClosedError:
        pass
    finally:
        sockets.pop(name, None)


def loop_in_thread(port: str):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(websockets.serve(handler, global_backend_url, port))  # type: ignore
    loop.run_forever()


# async is not needed, keeping for backwards compatibility
async def init(frontend_url="http://interp-tools.redwoodresearch.org", port=6789, backend_url="localhost"):
    global thread, global_frontend_url, global_backend_url, global_port
    if str(port) == global_port:
        print(f"Composable UI server already running on localhost:{port} in this Python process")
        return
    if is_port_in_use(int(port)):
        print(
            f"Failed to start Composable UI server: port {port} already in use. Maybe you have another instance running?"
        )
        return
    global_frontend_url = frontend_url
    port = str(port)
    global_port = port
    global_backend_url = backend_url
    t = threading.Thread(target=loop_in_thread, args=(port,))
    t.start()
    thread = t
    print(f"Composable UI initialized on localhost:{port}!")


def html_link(link) -> HTML:
    return HTML(f'<h1><a href="{link}" target="_blank">Link</a><script>window.open({link},"_blank")</script></h1>')


MYPY = False
if MYPY:
    from interp.circuit.circuits_very_named_tensor import CircuitsVeryNamedTensor
    from interp.ui.very_named_tensor import VeryNamedTensor as oldVeryNamedTensor


async def show_tensors(
    *lvnts: Union[
        LazyVeryNamedTensor, VeryNamedTensor, rCircuitsVeryNamedTensor, CircuitsVeryNamedTensor, oldVeryNamedTensor
    ],
    name="untitled",
    show_tensors: Optional[list[int]] = None,
):
    vnt_types = (LazyVeryNamedTensor, rCircuitsVeryNamedTensor)
    try:
        from interp.circuit.circuits_very_named_tensor import CircuitsVeryNamedTensor

        vnt_types = vnt_types + (CircuitsVeryNamedTensor,)  # type: ignore
    except ImportError:
        pass

    lvnts = tuple(lvnt.to_lvnt() if not isinstance(lvnt, vnt_types) else lvnt for lvnt in lvnts)  # type: ignore
    with lock:
        did_exist = name in served_things
        served_things[name] = lvnts
    if did_exist:
        encoded = msgpack_encode({"kind": "nameStartup", "data": lvnts})
        try:
            await sockets[name].send(encoded)  # type: ignore
        except:
            pass
    target = "alltensors" if show_tensors is not None else "tensors"
    which_tensors = "" if show_tensors is None else "&showcompact=" + urllib.parse.quote(str(show_tensors))
    link = f"{global_frontend_url}/#/{target}/{name}?port={global_port}&url={global_backend_url}{which_tensors}"
    print(link)
    return display(html_link(link))


async def show_fns(*lvnt_makers, name="untitled"):
    with lock:
        did_exist = name in served_things
        served_things[name] = lvnt_makers
    link = f"{global_frontend_url}/#/functions/{name}/?port={global_port}&url={global_backend_url}"
    if did_exist:
        await sockets[name].send(msgpack_encode({"kind": "nameStartup", "data": lvnt_makers}))
    return display(html_link(link))


async def show_attribution(model, params, tokenizer, name="untitled"):
    from interp.ui.attribution_backend import AttributionBackend

    func = partial(AttributionBackend, model, params, tokenizer)
    with lock:
        did_exist = name in served_things
        served_things[name] = func
    if did_exist:
        await sockets[name].send(msgpack_encode({"kind": "nameStartup", "data": func}))
    link = f"{global_frontend_url}/#/attribution/{name}/?port={global_port}&url={global_backend_url}"
    return display(html_link(link))


def is_port_in_use(port: int) -> bool:
    # https://stackoverflow.com/a/52872579/4642943
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0



import pytest
import torch

import rust_circuit.optional as op
from interp.circuit.circuit import MemoizedFn
from interp.circuit.circuit_utils import evaluate_fn
from interp.circuit.computational_node import UnaryRearrange
from interp.circuit.constant import ArrayConstant
from rust_circuit import Array as rArray
from rust_circuit import Rearrange as rRearrange
from rust_circuit import RearrangeSpec as rRearrangeSpec


def test_repeat_at_axes():
    for (t, axes, counts) in [
        (torch.randn(()), (), ()),
        (torch.randn(()), (0,), (3,)),
        (torch.randn(()), (0,), (1,)),
        (torch.randn(()), (0,), None),
        (torch.randn(()), (0, 1, 2), None),
        (torch.randn(()), (0, 1, 2), (2, 1, 3)),
        (torch.randn((3,)), (), ()),
        (torch.randn((3,)), (0,), (2,)),
        (torch.randn((3,)), (0,), None),
        (torch.randn((3,)), (1,), (2,)),
        (torch.randn((3,)), (1,), None),
        (torch.randn((3,)), (0, 2), (2, 4)),
        (torch.randn((3,)), (0, 2), None),
        (torch.randn(3, 4, 2), (1, 2, 3, 5), None),
        (torch.randn(3, 4, 2), (1, 2, 3, 5), (3, 5, 1, 2)),
        (torch.randn(3, 4, 2), (), ()),
    ]:
        basic_r_args = [(axes, counts)]
        if len(axes) == 1:
            basic_r_args.append((axes[0], op.map(counts, lambda x: x[0])))

        new_py_node = UnaryRearrange.repeat_at_axes(ArrayConstant(t), axes, counts)
        for r_axes, r_counts in basic_r_args:
            new_rs_node = rRearrange(rArray(t), rRearrangeSpec.expand_at_axes(t.ndim, r_axes, r_counts))
            assert new_rs_node.shape == new_py_node.shape
            torch.testing.assert_close(new_rs_node.evaluate(), MemoizedFn(evaluate_fn())(new_py_node))

        new_py_node = UnaryRearrange.unsqueeze(ArrayConstant(t), axes)
        for r_axes, _ in basic_r_args:
            new_rs_node = rRearrange(rArray(t), rRearrangeSpec.unsqueeze(t.ndim, r_axes))
            assert new_rs_node.shape == new_py_node.shape
            torch.testing.assert_close(new_rs_node.evaluate(), MemoizedFn(evaluate_fn())(new_py_node))


def test_repeat_at_axes_fail():
    for (ndim, axes, counts) in [
        (0, (2,), ()),
        (0, (0,), (-1)),
        (3, (1, 2, 3, 5, 10), None),
        (3, (1, 2, 3, 5, 1000), None),
        (1000, (1, 2, 3, 5, 1000), None),
    ]:
        with pytest.raises((ValueError, TypeError)):
            rRearrangeSpec.expand_at_axes(ndim, axes, counts)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Sequence, Literal

import uuid
from dataclasses import dataclass
from uuid import UUID

import torch

import rust_circuit as rc
import rust_circuit.index_util.internal.ops as ops
import rust_circuit.index_util.internal.parse as parse
import rust_circuit.index_util.internal.resolve as resolve


class Gather(rc.GeneralFunctionSpecBase):
    """Get elements from a 'source' tensor using indices from a 'position' tensor.

    To create a 'Gather' node, use the 'Gather.new' method. The 'Gather.new' method takes a string specifying its
    behavior in a little programming language, similar to 'Einsum'. The program string must always be of the form:

        out[i0, i1, ..., iN] = src[..., pos[...], ...]

    where 'i0, i1, ..., iN' are the indices of the output tensor, 'src' is the name of the source tensor, and 'pos' is
    the name of the position tensor. The 'out' array must always be named 'out', but 'src' and 'pos' can be named
    whatever you want; their names are determined based on the keys of the 'args' dictionary you pass to 'Gather.new'.
    The index expressions used in the square brackets after 'src' and 'pos' can be either the index variables 'i0, i1,
    ..., iN', or constant integers.

    The 'out_shape' argument to 'Gather.new' is optional. If 'out_shape" is not specified, the output shape is inferred
    from the program string and the shapes of the input tensors. If 'out_shape' is specified, it must be a list of
    integers or 'None'. If an element of the list is 'None', the corresponding dimension of the output shape is inferred
    from the program string.

    'Gather' supports arbitrary batching. If any of the tensors have a shape with rank greater than what is implied by
    the program string, any extra leading dimensions are treated as batch dimensions.

    Examples:

        # Get the activations for a different token position for each sequence in a batch:
        #  'activations' has shape [batch_size, seq_len, num_features]
        #  'token_positions' has shape [batch_size] and contains integers in the range [0, seq_len)
        Gather.new(
            "out[i, j] = activations[i, token_positions[i], j]",
            {"activations": activations, "token_positions": token_positions},
        )

        # The previous example can also be done with implicit batching!
        Gather.new(
            "out[j] = activations[token_positions, j]", # equivalent to previous example
            {"activations": activations, "token_positions": token_positions},
        )

        # Get the difference in logit values between two reference vocabulary tokens for each sequence in a batch:
        #   'logits' has shape [batch_size, vocab_size]
        #   'logit_indices' has shape [batch_size, 2] and contains integers in the range [0, vocab_size).
        logits_token0 = Gather.new(
            "out = logits[logit_indices[0]]", # uses implicit batching
            {"logits": logits, "logit_indices": logit_indices},
        )
        logits_token1 = Gather.new(
            "out = logits[logit_indices[1]]", # uses implicit batching
            {"logits": logits, "logit_indices": logit_indices},
        )
        diff = Add.minus(logits_token0, logits_token1)

        # Indices are allowed to repeat:
        Gather.new(
            "out[i, j] = foo[i, i, bar[i, j], j]", # allowed {"foo": foo, "bar": bar},
        )

        # Indices don't need to appear in the same order as the output:
        Gather.new(
            "out[i, j, k] = foo[k, bar[j, k, i], i, j], # allowed {"foo": foo, "bar": bar},
        )

        # When the output shape can't be inferred, you can specify the missing dimensions explicitly:
        Gather.new(
            "out[i, j] = foo[bar[j]]", {"foo": foo, "bar": bar}, out_shape=[20, None],
        )
    """

    _spec: resolve.GatherSpec
    _plan: ops.GatherPlan
    _orig_out_shape: Optional[list[Optional[int]]]

    def __init__(self, s: str, *args: str, out_shape: Optional[Sequence[Optional[int]]] = None) -> None:
        super().__init__()
        out_shape_list = list(out_shape) if out_shape is not None else None
        self._spec = resolve.GatherSpec.parse(s, args)
        self._plan = ops.plan_gather(self._spec)
        self._orig_out_shape = out_shape_list

    @property
    def name(self) -> str:
        out_shape_str = f", out_shape={self._orig_out_shape!r}" if self._orig_out_shape is not None else ""
        return f"Gather({str(self._spec)!r}, {self._spec.src_name!r}, {self._spec.pos_name!r}{out_shape_str})"

    def compute_hash_bytes(self) -> bytes:
        b = bytearray(uuid.UUID("7815ae94-c2bf-4e9b-a1a1-d2c96767c9a8").bytes)

        def extend_str(s: str) -> None:
            b.extend(len(s).to_bytes(8, "little"))
            b.extend(s.encode("utf-8"))

        extend_str(str(self._spec))
        extend_str(self._spec.src_name)
        extend_str(self._spec.pos_name)
        extend_str(str(self._orig_out_shape))

        return bytes(b)

    def get_shape_info(self, *shapes: rc.Shape) -> rc.GeneralFunctionShapeInfo:
        src_shape, pos_shape = shapes
        inferred = ops.infer_gather_shapes(
            self._plan, self._orig_out_shape, list(src_shape), list(pos_shape), self._spec.src_name, self._spec.pos_name
        )
        return rc.GeneralFunctionShapeInfo(
            tuple(inferred.batch_shape + inferred.axis_sizes),
            num_non_batchable_output_dims=len(self._orig_out_shape)
            if self._orig_out_shape is not None
            else len(inferred.axis_sizes),
            input_batchability=[True, True],
        )

    def function(self, *inputs: torch.Tensor) -> torch.Tensor:
        src, pos = inputs
        return ops.execute_gather(self._plan, self._orig_out_shape, src, pos, self._spec.src_name, self._spec.pos_name)

    def get_device_dtype_override(self, *device_dtypes: rc.TorchDeviceDtypeOp) -> Optional[rc.TorchDeviceDtypeOp]:
        src_dtype, pos_dtype = device_dtypes
        assert pos_dtype.dtype == "int64"
        return src_dtype

    @classmethod
    def new(  # type: ignore[override]
        cls,
        spec: str,
        args: dict[str, rc.Circuit],
        out_shape: Optional[Sequence[Optional[int]]] = None,
        name: Optional[str] = None,
    ) -> rc.Circuit:
        func_spec = cls(spec, *args.keys(), out_shape=out_shape)
        src_name = func_spec._spec.src_name
        pos_name = func_spec._spec.pos_name
        return rc.GeneralFunction(args[src_name], args[pos_name], spec=func_spec, name=name)


class Scatter(rc.GeneralFunctionSpecBase):
    """Write to selected elements of a 'destination' tensor using values from a 'source' tensor and indices from a
    'position' tensor.

    This operation is non-destructive; the destination tensor is not modified. Instead, a new tensor is created with the
    same shape as the destination tensor, and with the selected elements changed.

    To create a 'Scatter' node, use the 'Scatter.new' method. Like 'Gather.new', this method takes a string specifying
    the operation to be performed in a little programming language. The string must always be of the form:

        dst[..., pos[...], ...] <- src[...]

    where 'dst' is the name of the destination tensor, 'src' is the name of the source tensor, and 'pos' is the name of
    the position tensor. The 'dst', 'src', and 'pos' names can be chosen arbitrarily, but they must be consistent with
    the names used in the 'args' dictionary passed to 'Scatter.new'. The index expressions used in the square brackets after 'dst', 'src', and 'pos' can be either index variables or
    constant integers.

    'Scatter' supports arbitrary batching. If any of the tensors have a shape with rank greater than what is implied by
    the program string, any extra leading dimensions are treated as batch dimensions.

    The 'reduce' argument to 'Scatter.new' specifies how the source values are combined with the destination values. If
    'reduce' is 'None', the source values are written directly to the destination tensor and replace the old values. If
    'reduce' is 'add' or 'multiply', the source values are reduced with the destination values using the specified
    operation.

    Every index variable which appears on the right hand side of the formula must also appear on the left hand side.
    Moreover, due to a limitation of the underlying PyTorch 'scatter' operation, all but at most one of the variables
    appearing in the formula must appear as a direct index on 'dst' (see below for an example of this restriction).

    Examples:

        # Patch the activations from one batch of sequences into another, at different token position for each sequence
        # in the batch:

        #   'activations1' is the destination tensor, with shape [batch_size, seq_len, num_features]
        #   'activations2' is the source tensor, also with shape [batch_size, seq_len, num_features]
        #   'positions' is the position tensor, with shape [batch_size] and values in the range [0, seq_len)

        # First use a 'Gather' to extract the activations at the specified positions
        extracted = Gather.new(
            "out[i, j] = activations2[i, positions[i], j]",
            {"activations2": activations2, "positions": positions},
        )

        # Now, use a 'Scatter' to write the activations at the specified positions
        activations1_patched = Scatter.new(
            "activations1[i, positions[i], j] <- extracted[i, j]",
            {"activations1": activations1, "extracted": extracted},
        )

        # The same operation can be performed with implicit batching in both the gather and the scatter:

        extracted = Gather.new(
            "out[i] = activations2[positions, i]", # uses implicit batching
            {"activations2": activations2, "positions": positions},
        )
        activations1_patched = Scatter.new(
            "activations1[positions, i] <- extracted[i]", # uses implicit batching
            {"activations1": activations1, "extracted": extracted},
        )

        # You are allowed to repeat and reorder indices:
        scatter = Scatter.new(
            "foo[i, i, baz[i, j, i], k] <- bar[k, j, i]", # OK
            {"foo": foo, "bar": bar, "baz": baz},
        )

        # You can use constant integers to index into tensors:
        scatter = Scatter.new(
            "foo[i, 0, baz[i, 2], 1] <- bar[4, i]", # OK
            {"foo": foo, "bar": bar, "baz": baz},
        )

        # Every variable appearing on the right hand side must also appear on the left hand side:
        scatter = Scatter.new(
            "foo[i, baz[j]] <- bar[i, j, k]", # ERROR: 'k' does not appear on the left hand side
            {"foo": foo, "bar": bar, "baz": baz},
        )

        # All but at most one of the variables appearing in the formula must appear as a direct index on 'dst':
        scatter = Scatter.new(
            # ERROR: there are two variables which do not appear as direct indices on 'dst': 'j' and 'k'
            "foo[i, baz[j, k]] <- bar[i, j, k]",
            {"foo": foo, "bar": bar, "baz": baz},
        )
    """

    _spec: resolve.ScatterSpec
    _plan: ops.ScatterPlan

    def __init__(self, s: str, *args: str, reduce: Optional[Literal["add", "multiply"]] = None) -> None:
        super().__init__()
        self._spec = resolve.ScatterSpec.parse(s, args, reduce=reduce)
        self._plan = ops.plan_scatter(self._spec)

    @property
    def name(self) -> str:
        return (
            f"Scatter({str(self._spec)!r}, "
            + f"{self._spec.dst_name!r}, {self._spec.src_name!r}, {self._spec.pos_name!r}, "
            + f"reduce={self._spec.reduce!r})"
        )

    def compute_hash_bytes(self) -> bytes:
        b = bytearray(uuid.UUID("b3b3c3c3-1b2d-4c9b-9f7a-0f3e8b3d3b3b").bytes)

        def extend_str(s: str) -> None:
            b.extend(len(s).to_bytes(8, "little"))
            b.extend(s.encode("utf-8"))

        extend_str(str(self._spec))
        extend_str(self._spec.dst_name)
        extend_str(self._spec.src_name)
        extend_str(self._spec.pos_name)
        extend_str(str(self._spec.reduce))

        return bytes(b)

    def get_shape_info(self, *shapes: rc.Shape) -> rc.GeneralFunctionShapeInfo:
        dst_shape, src_shape, pos_shape = shapes
        inferred, out_shape = ops.infer_scatter_shapes(
            self._plan,
            list(dst_shape),
            list(src_shape),
            list(pos_shape),
            self._spec.dst_name,
            self._spec.src_name,
            self._spec.pos_name,
        )
        return rc.GeneralFunctionShapeInfo(
            tuple(out_shape),
            num_non_batchable_output_dims=len(out_shape) - len(inferred.batch_shape),
            input_batchability=[True, True, True],
        )

    def function(self, *inputs: torch.Tensor) -> torch.Tensor:
        dst, src, pos = inputs
        return ops.execute_scatter(
            self._plan, dst, src, pos, self._spec.dst_name, self._spec.src_name, self._spec.pos_name
        )

    def get_device_dtype_override(self, *device_dtypes: rc.TorchDeviceDtypeOp) -> Optional[rc.TorchDeviceDtypeOp]:
        dst_dtype, src_dtype, pos_dtype = device_dtypes
        assert pos_dtype.dtype == "int64"
        return dst_dtype

    @classmethod
    def new(  # type: ignore[override]
        cls,
        spec: str,
        args: dict[str, rc.Circuit],
        reduce: Optional[Literal["add", "multiply"]] = None,
        name: Optional[str] = None,
    ) -> rc.Circuit:
        func_spec = cls(spec, *args.keys(), reduce=reduce)
        dst_name = func_spec._spec.dst_name
        src_name = func_spec._spec.src_name
        pos_name = func_spec._spec.pos_name
        return rc.GeneralFunction(args[dst_name], args[src_name], args[pos_name], spec=func_spec, name=name)



# this file is included with `include_str!()` in py_types.rs

import atexit
import functools
import hashlib
from typing import Dict, Tuple

import numpy as np
import torch
from blake3 import blake3

PER_BLOCK_THREAD_COUNT = 128


@functools.lru_cache(maxsize=None)
def get_cuda_kernels():
    import pycuda.autoprimaryctx as apc  # type: ignore

    ctx = apc.context
    # atexit.unregister(apc._finish_up)

    import pycuda.driver as cuda  # type: ignore
    from pycuda.compiler import SourceModule  # type: ignore

    # cuda hash by Peter Schmidt-Nielsen
    mod = SourceModule(
        """
    #include <cstdint>
    static const int PER_BLOCK_THREAD_COUNT = %i;
    // Constants are chunks digits of pi, but |1 to be odd.
    __device__ static const uint32_t ODD_CONSTANT32[8]= { 0x243f6a89ul, 0x85a308d3ul, 0x13198a2ful, 0x03707345ul,
    0xa4093823ul, 0x299f31d1ul, 0x082efa99ul, 0xec4e6c89ul,
    };
    #define MIX_STATE(s) do { \
        s[1] ^= s[0] >> 15; s[1] *= ODD_CONSTANT32[0]; \
        s[2] ^= s[1] >> 15; s[2] *= ODD_CONSTANT32[1]; \
        s[3] ^= s[2] >> 15; s[3] *= ODD_CONSTANT32[2]; \
        s[4] ^= s[3] >> 15; s[4] *= ODD_CONSTANT32[3]; \
        s[5] ^= s[4] >> 15; s[5] *= ODD_CONSTANT32[4]; \
        s[6] ^= s[5] >> 15; s[6] *= ODD_CONSTANT32[5]; \
        s[7] ^= s[6] >> 15; s[7] *= ODD_CONSTANT32[6]; \
        s[0] ^= s[7] >> 15; s[0] *= ODD_CONSTANT32[7]; \
    } while(0)
    __global__ void _gpuhash(
    const uint8_t* data,
    int64_t length,
    uint32_t* block_hash_buffer
    ) {
    extern __shared__ uint32_t sdata[];
    // Each CUDA block hashes one contiguous sub-block of the data -- figure out ours.
    int64_t per_block_length = (length + (int64_t) gridDim.x - 1) / (int64_t) gridDim.x;
    int64_t our_limit = min(length, (1 + (int64_t) blockIdx.x) * per_block_length);
    uint32_t state[8] = { blockIdx.x, threadIdx.x, 1, 2, 3, 4, 5, 6 };
    for (int i = 0; i < 8; i++)
        MIX_STATE(state);
    // Hash the entire sub-block, with each thread having a 256-bit state.
    for (size_t i = threadIdx.x + blockIdx.x * per_block_length; i < our_limit; i += PER_BLOCK_THREAD_COUNT) {
        state[0] ^= data[i];
        MIX_STATE(state);
    }
    for (int i = 0; i < 8; i++)
        sdata[8 * threadIdx.x + i] = state[i];
    __syncthreads();
    // Reduce the individual thread hashes down into a single 256-bit block hash.
    #define REDUCE(n) \
        if (PER_BLOCK_THREAD_COUNT >= 2 * n) { \
        if (threadIdx.x < n) \
            for (int i = 0; i < 8; i++) \
                sdata[8 * threadIdx.x + i] ^= sdata[8 * (threadIdx.x + n) + i] * ODD_CONSTANT32[7 - i]; \
        __syncthreads(); \
        }
    static_assert(PER_BLOCK_THREAD_COUNT <= 1024, "block_size too large");
    REDUCE(512) REDUCE(256) REDUCE(128) REDUCE(64) REDUCE(32)
    REDUCE(16)  REDUCE(8)   REDUCE(4)   REDUCE(2)  REDUCE(1)
    if (threadIdx.x < 8)
        block_hash_buffer[8 * blockIdx.x + threadIdx.x] = sdata[threadIdx.x];
    }
    __global__ void _gpuhash_finalize(
    uint32_t* block_hash_buffer,
    int length
    ) {
    uint32_t state[8] = {1, 2, 3, 4, 5, 6, 7, 8};
    // Absorb all of the 256-bit block hashes.
    for (int i = 0; i < length; i++) {
        state[0] ^= block_hash_buffer[i];
        for (int j = 0; j < 2; j++)
        MIX_STATE(state);
    }
    // Final mixing, and then output the 256-bit hash.
    for (int i = 0; i < 8; i++)
        MIX_STATE(state);
    block_hash_buffer[threadIdx.x] = state[threadIdx.x];
    }
    """
        % PER_BLOCK_THREAD_COUNT
    )

    _gpuhash = mod.get_function("_gpuhash")
    _gpuhash_finalize = mod.get_function("_gpuhash_finalize")
    return (_gpuhash, _gpuhash_finalize, ctx)


dtypes_to_bytes = {
    torch.float32: b"float32",
    torch.float64: b"float64",
    torch.float16: b"float16",
    torch.bool: b"bool---",
    torch.int8: b"int8---",
    torch.int16: b"int16--",
    torch.int32: b"int32--",
    torch.int64: b"int64--",
    torch.uint8: b"uint8--",
}


@functools.lru_cache(maxsize=None)
def int_tuple_to_bytes(s: Tuple[int, ...], b):
    for x in s:
        b += x.to_bytes(8, "big", signed=True)
    return b


import weakref

tensor_hash_cache: Dict[int, Tuple[weakref.ref[torch.Tensor], bytes]] = {}


def hash_tensor(tensor: torch.Tensor) -> bytes:
    if isinstance(tensor, torch.nn.Parameter):
        raise TypeError(
            f"Circuits assumes tensors are immutable, but this tensor " + "is a Parameter! You should call .clone()."
        )

    if (
        (cached := tensor_hash_cache.get(id(tensor))) is not None
        and (cached_tensor := cached[0]()) is not None
        and cached_tensor is tensor
    ):
        # print("hit hash cache!")
        return cached[1]
    # print(f"didn't hit hash cache {id(tensor)=}")
    hash_out = hash_tensor_impl(tensor)
    tensor_hash_cache[id(tensor)] = (weakref.ref(tensor), hash_out)
    return hash_out


def hash_tensor_impl(tensor):
    m = blake3(max_threads=blake3.AUTO)
    m.update(b"c3c5ce3e-7d06-4939-8340-2a7e6b2984b4")
    m.update(int_tuple_to_bytes(tensor.shape, b""))
    m.update(b"fb782993-5fe7-43ac-acc2-d57fb429c2fc")
    m.update(dtypes_to_bytes[tensor.dtype])
    m.update(str(tensor.device).encode("utf-8"))
    m.update(b"9430f5c7-3c58-4339-b863-5449a4d74ee6")
    if str(tensor.device) == "cpu":
        m.update(tensor.detach().numpy().tobytes())  # this copies, could be optimized more
    else:
        if True:
            m.update(tensor.detach().cpu().numpy().tobytes())
        else:
            try:
                m.update(hash_tensor_contents_gpu(tensor))
            except Exception as e:
                print("gpu hashing failed, maybe because cuda version < 11.4 or pycuda not installed")
                print(e)
                m.update(tensor.detach().cpu().numpy().tobytes())

    return m.digest()[:32]


def pop_cuda_context():
    _, _, ctx = get_cuda_kernels()
    ctx.pop()


def hash_tensor_contents_gpu(tensor):
    # this makes pycuda use the default context, whereas the
    # default pycuda.autoinit uses a new context which clashes with pytorch

    tensor = tensor.contiguous()
    data_length = tensor.element_size() * tensor.nelement()
    block_count = 32
    if data_length > 1e6:
        block_count = 64
    if data_length > 1e7:
        block_count = 128
    if data_length > 1e8:
        block_count = 256
    block_hash_buffer = torch.zeros(size=(block_count * 8,), dtype=torch.int32, device=tensor.device)
    _gpuhash, _gpuhash_finalize, _ = get_cuda_kernels()
    _gpuhash(
        tensor,
        np.int64(data_length),
        block_hash_buffer,
        grid=(block_count, 1, 1),
        block=(PER_BLOCK_THREAD_COUNT, 1, 1),
        shared=32 * PER_BLOCK_THREAD_COUNT,
    )
    _gpuhash_finalize(
        block_hash_buffer,
        np.int32(len(block_hash_buffer)),
        grid=(1, 1, 1),
        block=(8, 1, 1),
    )
    b = block_hash_buffer[:8].cpu().numpy().tobytes()
    return b


if __name__ == "__main__":
    # tensor1 = torch.randn(10, 10, device="cuda:0")
    tensor1 = torch.randn(4_000_000_000, dtype=torch.float32, device="cuda")
    torch.cuda.synchronize()
    import time

    for _ in range(3):
        t = time.time()
        hash_tensor_contents_gpu(tensor1)
        torch.cuda.synchronize()
        print(time.time() - t)
